import inspect
import itertools
import os
import tempfile
from contextlib import contextmanager

from dagster import check


def script_relative_path(file_path):
    '''
    Useful for testing with local files. Use a path relative to where the
    test resides and this function will return the absolute path
    of that file. Otherwise it will be relative to script that
    ran the test
    '''
    # from http://bit.ly/2snyC6s

    check.str_param(file_path, 'file_path')
    scriptdir = inspect.stack()[1][1]
    return os.path.join(os.path.dirname(os.path.abspath(scriptdir)), file_path)


def _unlink_swallow_errors(path):
    check.str_param(path, 'path')
    try:
        os.unlink(path)
    except:  # pylint: disable=W0702
        pass


@contextmanager
def get_temp_file_name():
    temp_file_name = tempfile.mkstemp()[1]
    try:
        yield temp_file_name
    finally:
        _unlink_swallow_errors(temp_file_name)


@contextmanager
def get_temp_file_names(number):
    check.int_param(number, 'number')

    temp_file_names = list()
    for _ in itertools.repeat(None, number):
        temp_file_name = tempfile.mkstemp()[1]
        temp_file_names.append(temp_file_name)

    try:
        yield tuple(temp_file_names)
    finally:
        for temp_file_name in temp_file_names:
            _unlink_swallow_errors(temp_file_name)



from dagster import (
    check,
    ExpectationDefinition,
    InputDefinition,
    MaterializationDefinition,
    OutputDefinition,
    SourceDefinition,
)

# This is a collection point for older APIs that should no longer be "public"
# but still used in unit tests because I don't want to rewrite them all


def create_custom_source_input(
    name,
    source_fn,
    *,
    argument_def_dict=None,
    depends_on=None,
    expectations=None,
    source_type='CUSTOM'
):
    '''
    This function exist and is used a lot because separation of inputs and sources used to not
    exist so most of the unit tests in the systems were written without tha abstraction. So
    this exists as a bridge from the old api to the new api.
    '''
    return InputDefinition(
        name=name,
        sources=[
            SourceDefinition(
                source_type=source_type,
                source_fn=source_fn,
                argument_def_dict=argument_def_dict,
            )
        ],
        depends_on=depends_on,
        expectations=check.opt_list_param(
            expectations, 'expectations', of_type=ExpectationDefinition
        )
    )

def create_single_materialization_output(
    name, materialization_fn, argument_def_dict, expectations=None
):
    '''
    Similar to create_single_source_input this exists because a move in the primitive APIs.
    Materializations and outputs used to not be separate concepts so this is a compatability
    layer with the old api. Once the *new* api stabilizes this should be removed but it stays
    for now.
    '''
    return OutputDefinition(
        materializations=[
            MaterializationDefinition(
                name=name,
                materialization_fn=materialization_fn,
                argument_def_dict=argument_def_dict,
            )
        ],
        expectations=expectations
    )



import sqlalchemy as sa
import dagster.sqlalchemy_kernel as dagster_sa


def create_num_table(engine, num_table_name='num_table'):
    metadata = sa.MetaData(engine)

    table = sa.Table(
        num_table_name,
        metadata,
        sa.Column('num1', sa.Integer),
        sa.Column('num2', sa.Integer),
    )

    table.create()

    conn = engine.connect()

    conn.execute(f'''INSERT INTO {num_table_name} VALUES(1, 2)''')
    conn.execute(f'''INSERT INTO {num_table_name} VALUES(3, 4)''')


def in_mem_engine(num_table_name='num_table'):
    engine = sa.create_engine('sqlite://')
    create_num_table(engine, num_table_name)
    return engine


def in_mem_context(num_table_name='num_table'):
    return dagster_sa.create_sql_alchemy_context_from_engine(engine=in_mem_engine(num_table_name))



import pytest

from dagster import check, config
from dagster.core import types

from dagster.core.definitions import (
    SolidDefinition,
    ExpectationDefinition,
    ExpectationResult,
    ArgumentDefinition,
    OutputDefinition,
)

from dagster.core.execution import (
    output_single_solid,
    SolidExecutionResult,
    DagsterExecutionFailureReason,
    ExecutionContext,
    execute_single_solid,
)

from dagster.core.errors import DagsterExpectationFailedError

from dagster.utils.compatability import (
    create_custom_source_input, create_single_materialization_output
)


def create_test_context():
    return ExecutionContext()


def single_input_env(solid_name, input_name, args=None):
    check.str_param(solid_name, 'solid_name')
    check.str_param(input_name, 'input_name')
    args = check.opt_dict_param(args, 'args')
    return config.Environment(sources={solid_name: {input_name: config.Source('CUSTOM', args)}})


def test_execute_solid_no_args():
    some_input = create_custom_source_input(
        name='some_input',
        source_fn=lambda context, arg_dict: [{'data_key': 'data_value'}],
        argument_def_dict={}
    )

    def tranform_fn_inst(_context, args):
        args['some_input'][0]['data_key'] = 'new_value'
        return args['some_input']

    test_output = {}

    def materialization_fn_inst(context, arg_dict, data):
        assert isinstance(context, ExecutionContext)
        assert isinstance(arg_dict, dict)
        test_output['thedata'] = data

    custom_output = create_single_materialization_output(
        name='CUSTOM',
        materialization_fn=materialization_fn_inst,
        argument_def_dict={},
    )

    single_solid = SolidDefinition(
        name='some_node',
        inputs=[some_input],
        transform_fn=tranform_fn_inst,
        output=custom_output,
    )

    output_single_solid(
        create_test_context(),
        single_solid,
        environment=single_input_env('some_node', 'some_input'),
        name='CUSTOM',
        arg_dict={}
    )

    assert test_output['thedata'] == [{'data_key': 'new_value'}]


def create_single_dict_input(expectations=None):
    return create_custom_source_input(
        name='some_input',
        source_fn=lambda context, arg_dict: [{'key': arg_dict['str_arg']}],
        argument_def_dict={'str_arg' : ArgumentDefinition(types.String)},
        expectations=expectations or [],
    )


def create_noop_output(test_output, expectations=None):
    def set_test_output(context, arg_dict, output):
        assert isinstance(context, ExecutionContext)
        assert arg_dict == {}
        test_output['thedata'] = output

    return create_single_materialization_output(
        name='CUSTOM',
        materialization_fn=set_test_output,
        argument_def_dict={},
        expectations=expectations
    )


def test_hello_world():
    def transform_fn(context, args):
        assert isinstance(context, ExecutionContext)
        hello_world_input = args['hello_world_input']
        assert isinstance(hello_world_input, dict)
        hello_world_input['hello'] = 'world'
        return hello_world_input

    output_events = {}

    def materialization_fn(context, arg_dict, data):
        assert data['hello'] == 'world'
        assert isinstance(context, ExecutionContext)
        assert arg_dict == {}
        output_events['called'] = True

    hello_world = SolidDefinition(
        name='hello_world',
        inputs=[
            create_custom_source_input(
                name='hello_world_input',
                source_fn=lambda context, arg_dict: {},
                argument_def_dict={},
            )
        ],
        transform_fn=transform_fn,
        output=create_single_materialization_output(
            name='CUSTOM', materialization_fn=materialization_fn, argument_def_dict={}
        ),
    )

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=single_input_env('hello_world', 'hello_world_input'),
    )

    assert result.success

    assert result.transformed_value['hello'] == 'world'

    assert 'called' not in output_events

    output_result = output_single_solid(
        create_test_context(),
        hello_world,
        environment=single_input_env('hello_world', 'hello_world_input'),
        name='CUSTOM',
        arg_dict={}
    )

    if output_result.exception:
        raise output_result.exception

    assert output_result.success

    assert 'called' in output_events


def test_execute_solid_with_args():
    test_output = {}

    single_solid = SolidDefinition(
        name='some_node',
        inputs=[create_single_dict_input()],
        transform_fn=lambda context, args: args['some_input'],
        output=create_noop_output(test_output),
    )

    result = output_single_solid(
        create_test_context(),
        single_solid,
        environment=single_input_env('some_node', 'some_input', {'str_arg': 'an_input_arg'}),
        name='CUSTOM',
        arg_dict={},
    )

    if result.exception:
        raise result.exception

    assert result.success

    assert test_output['thedata'][0]['key'] == 'an_input_arg'


def test_execute_solid_with_failed_input_expectation_non_throwing():
    single_solid = create_input_failing_solid()

    solid_execution_result = output_single_solid(
        create_test_context(),
        single_solid,
        environment=single_input_env('some_node', 'some_input', {'str_arg': 'an_input_arg'}),
        name='CUSTOM',
        arg_dict={},
        throw_on_error=False,
    )

    assert isinstance(solid_execution_result, SolidExecutionResult)
    assert solid_execution_result.success is False
    assert solid_execution_result.reason == DagsterExecutionFailureReason.EXPECTATION_FAILURE


def test_execute_solid_with_failed_input_expectation_throwing():
    single_solid = create_input_failing_solid()

    with pytest.raises(DagsterExpectationFailedError):
        output_single_solid(
            create_test_context(),
            single_solid,
            environment=single_input_env('some_node', 'some_input', {'str_arg': 'an_input_arg'}),
            name='CUSTOM',
            arg_dict={},
        )

    with pytest.raises(DagsterExpectationFailedError):
        output_single_solid(
            create_test_context(),
            single_solid,
            environment=single_input_env('some_node', 'some_input', {'str_arg': 'an_input_arg'}),
            name='CUSTOM',
            arg_dict={},
        )


def create_input_failing_solid():
    test_output = {}

    def failing_expectation_fn(_context, _info, _some_input):
        return ExpectationResult(success=False)

    failing_expect = ExpectationDefinition(name='failing', expectation_fn=failing_expectation_fn)

    return SolidDefinition(
        name='some_node',
        inputs=[create_single_dict_input(expectations=[failing_expect])],
        transform_fn=lambda context, args: args['some_input'],
        output=create_noop_output(test_output),
    )


def test_execute_solid_with_failed_output_expectation_non_throwing():
    failing_solid = create_output_failing_solid()

    solid_execution_result = output_single_solid(
        create_test_context(),
        failing_solid,
        environment=single_input_env('some_node', 'some_input', {'str_arg': 'an_input_arg'}),
        name='CUSTOM',
        arg_dict={},
        throw_on_error=False
    )

    assert isinstance(solid_execution_result, SolidExecutionResult)
    assert solid_execution_result.success is False
    assert solid_execution_result.reason == DagsterExecutionFailureReason.EXPECTATION_FAILURE


def test_execute_solid_with_failed_output_expectation_throwing():
    failing_solid = create_output_failing_solid()

    with pytest.raises(DagsterExpectationFailedError):
        output_single_solid(
            create_test_context(),
            failing_solid,
            environment=single_input_env('some_node', 'some_input', {'str_arg': 'an_input_arg'}),
            name='CUSTOM',
            arg_dict={},
        )

    with pytest.raises(DagsterExpectationFailedError):
        output_single_solid(
            create_test_context(),
            failing_solid,
            environment=single_input_env('some_node', 'some_input', {'str_arg': 'an_input_arg'}),
            name='CUSTOM',
            arg_dict={},
        )


def _set_key_value(ddict, key, value):
    ddict[key] = value


def test_execute_solid_with_no_inputs():
    did_run_dict = {}
    no_args_solid = SolidDefinition(
        name='no_args_solid',
        inputs=[],
        transform_fn=lambda context, args: _set_key_value(did_run_dict, 'did_run', True),
        output=OutputDefinition(),
    )

    result = execute_single_solid(
        ExecutionContext(), no_args_solid, environment=config.Environment.empty()
    )

    assert result.success
    assert did_run_dict['did_run'] is True


def create_output_failing_solid():
    test_output = {}

    def failing_expectation_fn(_context, _info, _output):
        return ExpectationResult(success=False)

    output_expectation = ExpectationDefinition(
        name='output_failure', expectation_fn=failing_expectation_fn
    )

    return SolidDefinition(
        name='some_node',
        inputs=[create_single_dict_input()],
        transform_fn=lambda context, args: args['some_input'],
        output=create_noop_output(test_output, expectations=[output_expectation]),
    )



import dagster
from dagster import check
from dagster import config
from dagster.core.definitions import SolidDefinition
from dagster.utils.compatability import create_custom_source_input


def _set_key_value(ddict, key, value):
    ddict[key] = value
    return value


def test_execute_solid_with_dep_only_inputs_no_api():
    did_run_dict = {}

    step_one_solid = SolidDefinition(
        name='step_one_solid',
        inputs=[],
        transform_fn=lambda context, args: _set_key_value(did_run_dict, 'step_one', True),
        output=dagster.OutputDefinition(),
    )

    only_dep_input = create_custom_source_input(
        name='step_one_solid',
        source_fn=lambda arg_dict: check.not_implemented('should not get here'),
        argument_def_dict={},
        depends_on=step_one_solid
    )

    step_two_solid = SolidDefinition(
        name='step_two_solid',
        inputs=[only_dep_input],
        transform_fn=lambda context, args: _set_key_value(did_run_dict, 'step_two', True),
        output=dagster.OutputDefinition(),
    )

    pipeline = dagster.PipelineDefinition(solids=[step_one_solid, step_two_solid])

    # from dagster.utils import logging

    pipeline_result = dagster.execute_pipeline(pipeline, environment=config.Environment.empty())

    assert pipeline_result.success

    for result in pipeline_result.result_list:
        assert result.success

    assert did_run_dict['step_one'] is True
    assert did_run_dict['step_two'] is True


def test_execute_solid_with_dep_only_inputs_with_api():
    did_run_dict = {}

    step_one_solid = SolidDefinition(
        name='step_one_solid',
        inputs=[],
        transform_fn=lambda context, args: _set_key_value(did_run_dict, 'step_one', True),
        output=dagster.OutputDefinition(),
    )

    step_two_solid = SolidDefinition(
        name='step_two_solid',
        transform_fn=lambda context, args: _set_key_value(did_run_dict, 'step_two', True),
        inputs=[dagster.InputDefinition(step_one_solid.name, depends_on=step_one_solid)],
        output=dagster.OutputDefinition(),
    )

    pipeline = dagster.PipelineDefinition(solids=[step_one_solid, step_two_solid])

    pipeline_result = dagster.execute_pipeline(pipeline, environment=config.Environment.empty())

    for result in pipeline_result.result_list:
        assert result.success

    assert did_run_dict['step_one'] is True
    assert did_run_dict['step_two'] is True



import dagster
from dagster import config
from dagster.core.definitions import SolidDefinition
from dagster.utils.compatability import (
    create_custom_source_input, create_single_materialization_output
)
from dagster.core.execution import output_single_solid


def test_iterator_solid():
    def input_fn(_context, _arg_dict):
        yield 1
        yield 2

    some_input = create_custom_source_input(
        name='iter_numbers',
        source_fn=input_fn,
        argument_def_dict={},
    )

    def transform_fn(_context, args):
        for value in args['iter_numbers']:
            yield value + 1

    output_spot = {}

    def materialization_fn(_context, _arg_dict, data_iter):
        output_spot['list'] = list(data_iter)

        # in a real case we would iterate over
        # and stream to disk

    custom_output = create_single_materialization_output(
        name='CUSTOM',
        materialization_fn=materialization_fn,
        argument_def_dict={},
    )

    iterable_solid = SolidDefinition(
        name='some_node',
        inputs=[some_input],
        transform_fn=transform_fn,
        output=custom_output,
    )

    output_single_solid(
        dagster.ExecutionContext(),
        iterable_solid,
        environment=config.Environment(
            sources={'some_node': {
                'iter_numbers': config.Source('CUSTOM', {})
            }}
        ),
        name='CUSTOM',
        arg_dict={}
    )

    assert output_spot['list'] == [2, 3]



'''
Naming conventions:

For public functions:

execute_*

These represent functions which do purely in-memory compute. They will evaluate expectations
the core transform, and exercise all logging and metrics tracking (outside of outputs), but they
will not invoke *any* outputs (and their APIs don't allow the user to).

materialize_*

Materializations functions do execution but also allow the user to specify materializations,
which create artifacts that are discoverable by external systems (e.g. files, database
tables, and so on).


'''

# too many lines
# pylint: disable=C0302

from collections import (namedtuple, OrderedDict)
from contextlib import contextmanager
import copy
import sys

import six

from dagster import check, config

from dagster.utils.logging import (CompositeLogger, ERROR, get_formatted_stack_trace)
from dagster.utils.timing import time_execution_scope

from .definitions import (
    SolidDefinition, ExpectationResult, SourceDefinition,
    MaterializationDefinition, PipelineDefinition, PipelineContextDefinition
)

from .errors import (
    DagsterUserCodeExecutionError, DagsterTypeError, DagsterExecutionFailureReason,
    DagsterExpectationFailedError, DagsterInvariantViolationError
)

from .argument_handling import validate_args

Metric = namedtuple('Metric', 'context_dict metric_name value')


class ExecutionContext:
    '''
    A context object flowed through the entire scope of single execution of a
    pipeline of solids. This is used by both framework and user code to log
    messages and metrics. It also maintains a stack of context values so that
    logs, metrics, and any future reporting are reported with a minimal, consistent
    level of context so that developers do not have to repeatedly log well-known
    information (e.g. the name of the solid, the name of the pipeline, etc) when
    logging. Additionally tool author may add their own context values to assist
    reporting.


    resources is an arbitrary user-defined object that can be passed in
    by a user and then access during pipeline execution. This exists so that
    a user does not have to subclass ExecutionContext
    '''

    def __init__(self, loggers=None, log_level=ERROR, resources=None):
        self._logger = CompositeLogger(loggers=loggers, level=log_level)
        self._context_dict = OrderedDict()
        self._metrics = []
        self.log_level = log_level
        self.resources = resources

    def _maybe_quote(self, val):
        str_val = str(val)
        if ' ' in str_val:
            return '"{val}"'.format(val=str_val)
        return str_val

    def _kv_message(self, extra=None):
        extra = check.opt_dict_param(extra, 'extra')
        return ' '.join(
            [
                '{key}={value}'.format(key=key, value=self._maybe_quote(value))
                for key, value in [*self._context_dict.items(), *extra.items()]
            ]
        )

    def _log(self, method, msg, **kwargs):
        check.str_param(method, 'method')
        check.str_param(msg, 'msg')

        full_message = 'message="{message}" {kv_message}'.format(
            message=msg, kv_message=self._kv_message(kwargs)
        )

        log_props = copy.copy(self._context_dict)
        log_props['log_message'] = msg

        getattr(self._logger, method)(full_message, extra={**log_props, **kwargs})

    def debug(self, msg, **kwargs):
        return self._log('debug', msg, **kwargs)

    def info(self, msg, **kwargs):
        return self._log('info', msg, **kwargs)

    def warn(self, msg, **kwargs):
        return self._log('warn', msg, **kwargs)

    def error(self, msg, **kwargs):
        return self._log('error', msg, **kwargs)

    def critical(self, msg, **kwargs):
        return self._log('critical', msg, **kwargs)

    # FIXME: Actually make this work
    # def exception(self, e):
    #     check.inst_param(e, 'e', Exception)

    #     # this is pretty lame right. should embellish with more data (stack trace?)
    #     return self._log('error', str(e))

    @contextmanager
    def value(self, key, value):
        check.str_param(key, 'key')
        check.not_none_param(value, 'value')

        check.invariant(not key in self._context_dict, 'Should not be in context')

        self._context_dict[key] = value

        yield

        self._context_dict.pop(key)

    def metric(self, metric_name, value):
        check.str_param(metric_name, 'metric_name')
        check.not_none_param(value, 'value')

        keys = list(self._context_dict.keys())
        keys.append(metric_name)
        if isinstance(value, float):
            format_string = 'metric:{metric_name}={value:.3f} {kv_message}'
        else:
            format_string = 'metric:{metric_name}={value} {kv_message}'

        self._logger.info(
            format_string.format(
                metric_name=metric_name, value=value, kv_message=self._kv_message()
            ),
            extra=self._context_dict
        )

        self._metrics.append(
            Metric(
                context_dict=copy.copy(self._context_dict), metric_name=metric_name, value=value
            )
        )

    def _dict_covers(self, needle_dict, haystack_dict):
        for key, value in needle_dict.items():
            if not key in haystack_dict:
                return False
            if value != haystack_dict[key]:
                return False
        return True

    def metrics_covering_context(self, needle_dict):
        for metric in self._metrics:
            if self._dict_covers(needle_dict, metric.context_dict):
                yield metric

    def metrics_matching_context(self, needle_dict):
        for metric in self._metrics:
            if needle_dict == metric.context_dict:
                yield metric

class DagsterPipelineExecutionResult:
    def __init__(
        self,
        context,
        result_list,
    ):
        self.context = check.inst_param(context, 'context', ExecutionContext)
        self.result_list = check.list_param(
            result_list, 'result_list', of_type=SolidExecutionResult
        )

    @property
    def success(self):
        return all([result.success for result in self.result_list])

    def result_named(self, name):
        check.str_param(name, 'name')
        for result in self.result_list:
            if result.name == name:
                return result
        check.failed('Did not find result {name} in pipeline execution result'.format(name=name))


class SolidExecutionResult:
    '''
    A class to represent the result of the execution of a single solid. Pipeline
    commands return iterators or lists of these results.

    (TODO: explain the various error states)
    '''

    def __init__(
        self,
        success,
        solid,
        transformed_value,
        reason=None,
        exception=None,
        input_expectation_results=None,
        output_expectation_results=None,
        context=None,
    ):
        self.success = check.bool_param(success, 'success')
        if not success:
            check.param_invariant(
                isinstance(reason, DagsterExecutionFailureReason), 'reason',
                'Must provide a reason is result is a failure'
            )
        self.transformed_value = transformed_value
        self.solid = check.inst_param(solid, 'solid', SolidDefinition)
        self.reason = reason
        self.exception = check.opt_inst_param(exception, 'exception', Exception)

        self.input_expectation_results = check.opt_inst_param(
            input_expectation_results,
            'input_expectation_results',
            InputExpectationResults
        )

        self.output_expectation_results = check.opt_inst_param(
            output_expectation_results,
            'output_expectation_results',
            OutputExpectationResults,
        )

        if reason == DagsterExecutionFailureReason.USER_CODE_ERROR:
            check.inst(exception, DagsterUserCodeExecutionError)
            self.user_exception = exception.user_exception
        else:
            self.user_exception = None

        self.context = context

    def reraise_user_error(self):
        check.invariant(self.reason == DagsterExecutionFailureReason.USER_CODE_ERROR)
        check.inst(self.exception, DagsterUserCodeExecutionError)
        six.reraise(*self.exception.original_exc_info)

    @property
    def name(self):
        return self.solid.name

    def copy(self):
        ''' This must be used instead of copy.deepcopy() because exceptions cannot
        be deepcopied'''
        return SolidExecutionResult(
            success=self.success,
            solid=self.solid,
            transformed_value=copy.deepcopy(self.transformed_value),
            context=self.context,
            reason=self.reason,
            exception=self.exception,
            input_expectation_results=self.input_expectation_results.copy()
                    if self.input_expectation_results else None,
            output_expectation_results=self.output_expectation_results.copy()
                    if self.output_expectation_results else None,
        )

def copy_result_list(result_list):
    if result_list is None:
        return result_list

    return [result.copy() for result in result_list]


def copy_result_dict(result_dict):
    if result_dict is None:
        return None
    new_dict = {}
    for input_name, result in result_dict.items():
        new_dict[input_name] = result.copy()
    return new_dict

@contextmanager
def _user_code_error_boundary(context, msg, **kwargs):
    '''
    Wraps the execution of user-space code in an error boundary. This places a uniform
    policy around an user code invoked by the framework. This ensures that all user
    errors are wrapped in the SolidUserCodeExecutionError, and that the original stack
    trace of the user error is preserved, so that it can be reported without confusing
    framework code in the stack trace, if a tool author wishes to do so. This has
    been especially help in a notebooking context.
    '''
    check.inst_param(context, 'context', ExecutionContext)
    check.str_param(msg, 'msg')

    try:
        yield
    except Exception as e:
        stack_trace = get_formatted_stack_trace(e)
        context.error(str(e), stack_trace=stack_trace)
        raise DagsterUserCodeExecutionError(
            msg.format(**kwargs), e, user_exception=e, original_exc_info=sys.exc_info()
        )

def _read_source(context, source_definition, arg_dict):
    '''
    Check to ensure that the arguments to a particular input are valid, and then
    execute the input functions. Wraps that execution in appropriate logging, metrics tracking,
    and a user-code error boundary.
    '''
    check.inst_param(context, 'context', ExecutionContext)
    check.inst_param(source_definition, 'source_definition', SourceDefinition)
    check.dict_param(arg_dict, 'arg_dict', key_type=str)

    with context.value('source_type', source_definition.source_type):
        error_context_str = 'source type {source}'.format(source=source_definition.source_type)
        args_to_pass = validate_args(
            source_definition.argument_def_dict,
            arg_dict,
            error_context_str,
        )
        error_str = 'Error occured while loading source "{source_type}"'
        with _user_code_error_boundary(
            context,
            error_str,
            source_type=source_definition.source_type,
        ):
            context.info('Entering input implementation')

            with time_execution_scope() as timer_result:
                value = source_definition.source_fn(context, args_to_pass)

            context.metric('input_load_time_ms', timer_result.millis)

            return value


InputExpectationInfo = namedtuple('InputExpectionInfo', 'solid input_def expectation_def')

def _execute_input_expectation(context, info, value):
    '''
    Execute one user-specified input expectation on an input that has been instantiated in memory
    Wraps computation in an error boundary and performs all necessary logging and metrics tracking
    (TODO: actually log and track metrics!)
    '''
    check.inst_param(context, 'context', ExecutionContext)
    check.inst_param(info, 'info', InputExpectationInfo)

    error_str = 'Error occured while evaluation expectation "{expectation_name}" in input'
    with _user_code_error_boundary(context, error_str, expectation_name=info.expectation_def.name):
        expectation_result = info.expectation_def.expectation_fn(context, info, value)

    if not isinstance(expectation_result, ExpectationResult):
        raise DagsterInvariantViolationError(
            'Must return ExpectationResult from expectation function'
        )
    return expectation_result


OutputExpectationInfo = namedtuple('OutputExpectationInfo', 'solid expectation_def')

def _execute_output_expectation(context, info, transformed_value):
    '''
    Execute one user-specified output expectation on an instantiated result of the core transform.
    Wraps computation in an error boundary and performs all necessary logging and metrics tracking
    (TODO: actually log and track metrics!)
    '''
    check.inst_param(context, 'context', ExecutionContext)
    # check.inst_param(expectation_def, 'expectation_def', ExpectationDefinition)
    check.inst_param(info, 'info', OutputExpectationInfo)

    error_str = 'Error occured while evaluation expectation "{expectation_name}" in output'
    expectation_def = info.expectation_def
    with _user_code_error_boundary(context, error_str, expectation_name=expectation_def.name):
        expectation_result = expectation_def.expectation_fn(context, info, transformed_value)

    if not isinstance(expectation_result, ExpectationResult):

        raise DagsterInvariantViolationError(
            'Must return ExpectationResult from expectation function'
        )

    return expectation_result


def _execute_core_transform(context, solid_transform_fn, values_dict):
    '''
    Execute the user-specified transform for the solid. Wrap in an error boundary and do
    all relevant logging and metrics tracking
    '''
    check.inst_param(context, 'context', ExecutionContext)
    check.callable_param(solid_transform_fn, 'solid_transform_fn')
    check.dict_param(values_dict, 'values_dict', key_type=str)

    error_str = 'Error occured during core transform'
    with _user_code_error_boundary(context, error_str):
        with time_execution_scope() as timer_result:
            transformed_value = solid_transform_fn(context, values_dict)

        context.metric('core_transform_time_ms', timer_result.millis)

        return transformed_value


def _execute_materialization(context, materialiation_def, arg_dict, value):
    '''
    Execute a single output, calling into user-specified code. Check validity
    of arguments into the output, do appropriate loggina and metrics tracking, and
    actually execute the output function with an appropriate error boundary.
    '''
    check.inst_param(context, 'context', ExecutionContext)
    check.inst_param(materialiation_def, 'materialization', MaterializationDefinition)
    check.dict_param(arg_dict, 'arg_dict', key_type=str)

    expected_args = set(materialiation_def.argument_def_dict.keys())
    received_args = set(arg_dict.keys())

    if expected_args != received_args:
        raise DagsterTypeError(
            'Argument mismatch in output. Expected {expected} got {received}'.format(
                expected=repr(expected_args),
                received=repr(received_args),
            )
        )

    for arg_name, arg_value in arg_dict.items():
        arg_def_type = materialiation_def.argument_def_dict[arg_name]
        if not arg_def_type.dagster_type.is_python_valid_value(arg_value):
            raise DagsterTypeError(
                'Expected type {typename} for arg {arg_name} in output but got {arg_value}'.format(
                    typename=arg_def_type.dagster_type.name,
                    arg_name=arg_name,
                    arg_value=repr(arg_value),
                )
            )

    error_str = 'Error during execution of materialization'
    with _user_code_error_boundary(context, error_str):
        context.info('Entering materialization implementation')
        materialiation_def.materialization_fn(context, arg_dict, value)


class InputExpectationResult:
    def __init__(self, input_name, all_results):
        self.input_name = check.str_param(input_name, 'input_name')
        self.all_results = check.list_param(all_results, 'all_results', ExpectationResult)

    @property
    def fails(self):
        for result in self.all_results:
            if not result.success:
                yield result

    @property
    def passes(self):
        for result in self.all_results:
            if result.success:
                yield result

    def copy(self):
        return InputExpectationResult(
            input_name=self.input_name,
            all_results=copy_result_list(self.all_results),
        )

class InputExpectationResults:
    def __init__(self, result_dict):
        check.dict_param(
            result_dict,
            'result_dict',
            key_type=str,
            value_type=InputExpectationResult
        )

        all_passes = []
        all_fails = []
        for run_results in result_dict.values():
            all_passes.extend(run_results.passes)
            all_fails.extend(run_results.fails)

        self.all_passes = all_passes
        self.all_fails = all_fails
        self.result_dict = result_dict

    def copy(self):
        return InputExpectationResults(copy_result_dict(self.result_dict))

    @property
    def success(self):
        return not self.all_fails


def _execute_all_input_expectations(context, input_manager, solid, values_dict):
    check.inst_param(context, 'context', ExecutionContext)
    check.dict_param(values_dict, 'values_dict', key_type=str)

    result_dict = {}

    if not input_manager.evaluate_expectations:
        return InputExpectationResults(result_dict)

    for input_name in values_dict.keys():
        input_def = solid.input_def_named(input_name)
        value = values_dict[input_name]

        results = []

        for input_expectation_def in input_def.expectations:
            user_expectation_result = _execute_input_expectation(
                context, InputExpectationInfo(solid, input_def, input_expectation_def), value,
            )
            results.append(user_expectation_result)

        input_result = InputExpectationResult(
            input_name=input_name,
            all_results=results,
        )
        result_dict[input_name] = input_result

    return InputExpectationResults(result_dict)

class OutputExpectationResults:
    def __init__(self, results):
        self.results = check.list_param(results, 'results', ExpectationResult)

    @property
    def success(self):
        for result in self.results:
            if not result.success:
                return False
        return True

    def copy(self):
        return OutputExpectationResults(copy_result_list(self.results))

def _pipeline_solid_in_memory(context, input_manager, solid, transform_values_dict):
    '''
    Given inputs that are already in memory. Evaluation all inputs expectations,
    execute the core transform, and then evaluate all output expectations.

    This is the core of the solid execution that does not touch any extenralized state, whether
    it be inputs or outputs.
    '''
    check.inst_param(context, 'context', ExecutionContext)
    check.inst_param(solid, 'solid', SolidDefinition)
    check.dict_param(transform_values_dict, 'transform_values_dict', key_type=str)

    input_expectation_results = _execute_all_input_expectations(
        context,
        input_manager,
        solid,
        transform_values_dict
    )

    if not input_expectation_results.success:
        return SolidExecutionResult(
            success=False,
            transformed_value=None,
            solid=solid,
            context=context,
            reason=DagsterExecutionFailureReason.EXPECTATION_FAILURE,
            input_expectation_results=input_expectation_results
        )

    context.info('Executing core transform')

    transformed_value = _execute_core_transform(context, solid.transform_fn, transform_values_dict)

    if not solid.output.dagster_type.is_python_valid_value(transformed_value):
        raise DagsterInvariantViolationError(f'''Solid {solid.name} output {transformed_value}
            which does not match the type for Dagster Type {solid.output.dagster_type.name}''')

    if solid.output.output_callback:
        solid.output.output_callback(context, transformed_value)

    output_expectation_results = _execute_output_expectations(
        context,
        input_manager,
        solid,
        transformed_value,
    )

    if not output_expectation_results.success:
        return SolidExecutionResult(
            success=False,
            transformed_value=None,
            solid=solid,
            context=context,
            reason=DagsterExecutionFailureReason.EXPECTATION_FAILURE,
            input_expectation_results=input_expectation_results,
            output_expectation_results=output_expectation_results,
        )

    return SolidExecutionResult(
        success=True,
        transformed_value=transformed_value,
        solid=solid,
        context=context,
        input_expectation_results=input_expectation_results,
        output_expectation_results=output_expectation_results,
    )


def _execute_output_expectations(context, input_manager, solid, transformed_value):
    if not input_manager.evaluate_expectations:
        return OutputExpectationResults([])

    output_expectation_result_list = []

    for output_expectation_def in solid.output.expectations:
        info = OutputExpectationInfo(solid=solid, expectation_def=output_expectation_def)
        output_expectation_result = _execute_output_expectation(context, info, transformed_value)
        output_expectation_result_list.append(output_expectation_result)

    return OutputExpectationResults(results=output_expectation_result_list)


def _create_passthrough_context_definition(context):
    check.inst_param(context, 'context', ExecutionContext)
    context_definition = PipelineContextDefinition(
        argument_def_dict={},
        context_fn=lambda _pipeline, _args: context
    )
    return {'default': context_definition}


def execute_single_solid(context, solid, environment, throw_on_error=True):
    check.inst_param(context, 'context', ExecutionContext)
    check.inst_param(solid, 'solid', SolidDefinition)
    check.inst_param(environment, 'environment', config.Environment)
    check.bool_param(throw_on_error, 'throw_on_error')

    check.invariant(environment.execution.from_solids == [])
    check.invariant(environment.execution.through_solids == [])

    single_solid_environment = config.Environment(
        sources=environment.sources,
        materializations=environment.materializations,
        expectations=environment.expectations,
        context=environment.context,
        execution=config.Execution.single_solid(solid.name),
    )

    pipeline_result = execute_pipeline(
        PipelineDefinition(
            solids=[solid],
            context_definitions=_create_passthrough_context_definition(context),
        ),
        environment=single_solid_environment,
    )

    results = pipeline_result.result_list
    check.invariant(len(results) == 1, 'must be one result got ' + str(len(results)))
    return results[0]


def _do_throw_on_error(execution_result):
    check.inst_param(execution_result, 'execution_result', SolidExecutionResult)
    if not execution_result.success:
        if execution_result.reason == DagsterExecutionFailureReason.EXPECTATION_FAILURE:
            raise DagsterExpectationFailedError(execution_result)
        elif execution_result.reason == DagsterExecutionFailureReason.USER_CODE_ERROR:
            execution_result.reraise_user_error()

        check.invariant(execution_result.exception)
        raise execution_result.exception


def output_single_solid(
    context,
    solid,
    environment,
    name,
    arg_dict,
    throw_on_error=True,
):
    check.inst_param(context, 'context', ExecutionContext)
    check.inst_param(solid, 'solid', SolidDefinition)
    check.inst_param(environment, 'environment', config.Environment)
    check.str_param(name, 'name')
    check.dict_param(arg_dict, 'arg_dict', key_type=str)
    check.bool_param(throw_on_error, 'throw_on_error')


    results = list(
        execute_pipeline_iterator(
            PipelineDefinition(
                solids=[solid],
                context_definitions=_create_passthrough_context_definition(context),
            ),
            environment=config.Environment(
                context=environment.context,
                sources=environment.sources,
                materializations=[
                    config.Materialization(
                        solid=solid.name, name=name, args=arg_dict
                    )
                ],
            ),
        )
    )

    check.invariant(len(results) == 1, 'must be one result got ' + str(len(results)))

    execution_result = results[0]

    check.invariant(execution_result.name == solid.name)

    if throw_on_error:
        _do_throw_on_error(execution_result)

    return execution_result


def _gather_input_values(context, solid, input_manager):
    check.inst_param(context, 'context', ExecutionContext)
    check.inst_param(solid, 'solid', SolidDefinition)
    check.inst_param(input_manager, 'input_manager', InputManager)

    context.info('About to instantiate and gather all inputs')

    input_values = {}
    for input_def in solid.inputs:
        with context.value('input', input_def.name):
            input_value = input_manager.get_input_value(context, solid, input_def)

            if not input_def.dagster_type.is_python_valid_value(input_value):
                raise DagsterInvariantViolationError(f'''Solid {solid.name} input {input_def.name}
                   received value {input_value} which does not match the type for Dagster type
                   {input_def.dagster_type.name}'''
                )

            input_values[input_def.name] = input_value
            if input_def.input_callback:
                input_def.input_callback(context, input_values[input_def.name])
    return input_values


def _execute_pipeline_solid_step(context, solid, input_manager):
    check.inst_param(context, 'context', ExecutionContext)
    check.inst_param(solid, 'solid', SolidDefinition)
    check.inst_param(input_manager, 'input_manager', InputManager)

    # The value produce by an inputs is potentially different per solid.
    # This is allowed so that two solids that do different instantiation of the
    # same exact input (e.g. the same file) don't have to create additional solids
    # to account for this.

    input_values = _gather_input_values(context, solid, input_manager)

    # This call does all input and output expectations, as well as the core transform
    execution_result = _pipeline_solid_in_memory(context, input_manager, solid, input_values)

    if not execution_result.success:
        return execution_result

    check.invariant(
        solid.name not in input_manager.intermediate_values,
        'should be not in intermediate values'
    )

    transformed_value = execution_result.transformed_value

    context.debug(
        'About to set {output} for {name}'.format(
            output=repr(transformed_value),
            name=solid.name,
        )
    )

    input_manager.intermediate_values[solid.name] = transformed_value

    return SolidExecutionResult(
        success=True,
        solid=solid,
        context=context,
        transformed_value=input_manager.intermediate_values[solid.name],
        input_expectation_results=execution_result.input_expectation_results,
        output_expectation_results=execution_result.output_expectation_results,
        exception=None
    )

class InputManager:
    def __init__(self):
        self.intermediate_values = {}

    @contextmanager
    def yield_context(self):
        check.not_implemented('must implement in subclass')

    @property
    def materializations(self):
        check.not_implemented('must implement in subclass')

    @property
    def evaluate_expectations(self):
        check.not_implemented('must implement in subclass')

    @property
    def from_solids(self):
        check.not_implemented('must implement in subclass')

    @property
    def through_solids(self):
        check.not_implemented('must implement in subclass')

    def get_input_value(self, context, solid, input_def):
        if input_def.depends_on and input_def.depends_on.name in self.intermediate_values:
            # grab value from dependency
            return self.intermediate_values[input_def.depends_on.name]
        else:
            # must get value from source
            return self._get_sourced_input_value(context, solid.name, input_def.name)

    def _get_sourced_input_value(self, _context, _solid_name, _input_name):
        check.not_implemented('must implement in subclass')


def _wrap_in_yield(thing):
    if isinstance(thing, ExecutionContext):
        def _wrap():
            yield thing

        return _wrap()

    return thing


class InMemoryInputManager(InputManager):
    def __init__(self, context, input_values, from_solids=None, through_solids=None):
        super().__init__()
        self.input_values = check.dict_param(input_values, 'input_values', key_type=str)
        self.context = check.inst_param(context, 'context', ExecutionContext)
        self._from_solids = check.opt_list_param(from_solids, from_solids, of_type=str)
        self._through_solids = check.opt_list_param(through_solids, through_solids, of_type=str)

    @property
    def from_solids(self):
        return self._from_solids

    @property
    def through_solids(self):
        return self._through_solids

    def _get_sourced_input_value(self, _context, _solid_name, input_name):
        return self.input_values[input_name]

    @contextmanager
    def yield_context(self):
        return _wrap_in_yield(self.context)

    @property
    def materializations(self):
        return []

    @property
    def evaluate_expectations(self):
        return True

def _validate_environment(environment, pipeline):
    for solid_name, input_configs in environment.sources.items():
        if not pipeline.has_solid(solid_name):
            raise DagsterInvariantViolationError(
                f'Solid "{solid_name} not found'
            )

        solid_inst = pipeline.solid_named(solid_name)

        for input_name, _source_configs in input_configs.items():
            if not solid_inst.has_input(input_name):
                raise DagsterInvariantViolationError(
                    f'Input "{input_name}" not found in the pipeline on solid "{solid_name}".' + \
                    f'Input must be one of {repr(pipeline.input_names)}'
                )

    context_name = environment.context.name

    if context_name not in pipeline.context_definitions:
        avaiable_context_keys = list(pipeline.context_definitions.keys())
        raise DagsterInvariantViolationError(f'Context {context_name} not found in ' + \
            f'pipeline definiton. Available contexts {repr(avaiable_context_keys)}'
        )


class EnvironmentInputManager(InputManager):
    def __init__(self, pipeline, environment):
        super().__init__()
        # This is not necessarily the best spot for these calls
        _validate_environment(environment, pipeline)
        self.pipeline = check.inst_param(pipeline, 'pipeline', PipelineDefinition)
        self.environment = check.inst_param(environment, 'environment', config.Environment)

    @property
    def from_solids(self):
        return self.environment.execution.from_solids

    @property
    def through_solids(self):
        return self.environment.execution.through_solids

    @contextmanager
    def yield_context(self):
        context_name = self.environment.context.name
        context_definition = self.pipeline.context_definitions[context_name]

        args_to_pass = validate_args(
            self.pipeline.context_definitions[context_name].argument_def_dict,
            self.environment.context.args,
            'pipeline {pipeline_name} context {context_name}'.format(
                pipeline_name=self.pipeline.name,
                context_name=context_name,
            )
        )

        thing = context_definition.context_fn(self.pipeline, args_to_pass)
        return _wrap_in_yield(thing)

    def _get_sourced_input_value(self, context, solid_name, input_name):
        source_config = self.environment.sources[solid_name][input_name]
        input_def = self.pipeline.get_input(solid_name, input_name)
        source_def = input_def.source_of_type(source_config.name)
        return _read_source(
            context, source_def, self.args_for_input(solid_name, input_def.name)
        )

    def args_for_input(self, solid_name, input_name):
        check.str_param(solid_name, 'solid_name')
        check.str_param(input_name, 'input_name')
        return self.environment.sources[solid_name][input_name].args

    @property
    def materializations(self):
        return self.environment.materializations

    @property
    def evaluate_expectations(self):
        return self.environment.expectations.evaluate

def execute_pipeline_iterator(pipeline, environment):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.inst_param(environment, 'enviroment', config.Environment)

    input_manager = EnvironmentInputManager(pipeline, environment)
    with input_manager.yield_context() as context:
        return _execute_pipeline_iterator(
            context,
            pipeline,
            EnvironmentInputManager(pipeline, environment)
        )

def execute_pipeline_iterator_in_memory(
    context,
    pipeline,
    input_values,
    *,
    from_solids=None,
    through_solids=None,
):
    check.opt_list_param(from_solids, 'from_solids', of_type=str)
    check.opt_list_param(through_solids, 'through_solids', of_type=str)
    return _execute_pipeline_iterator(
        context,
        pipeline,
        InMemoryInputManager(context, input_values, from_solids, through_solids),
    )

def _execute_pipeline_iterator(
    context,
    pipeline,
    input_manager,
):
    '''
    This is the core workhorse function of this module, iterating over the pipeline execution
    in topological order. This allow a tool consuming this API to execute a solid one at a time
    and then make decisions based upon the result.

    If you do not specify "through_solids" it executes all the solids specified entire pipeline.
    If through_solids is specified, it will stop executing once all of those solids in
    through_solids have been executed.

    If you want to actually output the results of the transform see output_pipeline_iterator

    execute_pipeline is the "synchronous" version of this function and returns a list of results
    once the entire pipeline has been executed.
    '''

    check.inst_param(context, 'context', ExecutionContext)
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.inst_param(input_manager, 'input_manager', InputManager)

    through_solids = input_manager.through_solids
    from_solids = input_manager.from_solids

    pipeline_context_value = pipeline.name if pipeline.name else 'unnamed'

    materialization_args = MaterializationArgs(pipeline, input_manager.materializations)

    if through_solids is None:
        through_solids = materialization_args.through_solids

    with context.value('pipeline', pipeline_context_value):
        input_manager = input_manager

        if not through_solids:
            through_solids = pipeline.all_sink_solids

        if not from_solids:
            all_deps = set()
            for through_solid in through_solids:
                all_deps.union(pipeline.solid_graph.transitive_dependencies_of(through_solid))

            from_solids = list(all_deps)

        # TODO provide meaningful error messages when the wrong inputs are provided for a particular
        # execution subgraph

        # for through_solid_name in through_solids:
        #     unprovided_inputs = pipeline.solid_graph.compute_unprovided_inputs(
        #         input_names=sourced_input_names, solid_name=through_solid_name
        #     )
        #     if unprovided_inputs:
        #         check.failed(
        #             'Failed to provide inputs {unprovided_inputs} for solid {name}'.format(
        #                 unprovided_inputs=unprovided_inputs, name=through_solid_name
        #             )
        #         )

        execution_graph = pipeline.solid_graph.create_execution_subgraph(
            from_solids, list(through_solids)
        )

        for solid in execution_graph.topological_solids:

            try:
                with context.value('solid', solid.name):
                    result = _execute_pipeline_solid_step(context, solid, input_manager)

                    if not result.success:
                        yield result
                        break

                    if not materialization_args.should_materialize(result.name):
                        yield result
                        continue

                    _execute_materializations(
                        context,
                        solid,
                        materialization_args.materializations_for_solid(solid.name),
                        result.transformed_value,
                    )

                    yield result

            except DagsterUserCodeExecutionError as see:
                yield SolidExecutionResult(
                    success=False,
                    reason=DagsterExecutionFailureReason.USER_CODE_ERROR,
                    solid=solid,
                    context=context,
                    transformed_value=None,
                    exception=see,
                )
                break

def _execute_materializations(
    context,
    solid,
    materializations,
    transformed_value,
):
    for materialization in materializations:
        arg_dict = materialization.args
        name = materialization.name
        with context.value('materialization_name', name):
            mat_def = solid.output.materialization_of_type(name)
            _execute_materialization(context, mat_def, arg_dict, transformed_value)

def execute_pipeline(
    pipeline,
    environment,
    *,
    throw_on_error=True,
):
    check.inst_param(environment, 'environment', config.Environment)
    return _execute_pipeline(
        pipeline,
        EnvironmentInputManager(pipeline, environment),
        throw_on_error,
    )

def execute_pipeline_in_memory(
    context,
    pipeline,
    *,
    input_values,
    from_solids=None,
    through_solids=None,
    throw_on_error=True,
):
    check.dict_param(input_values, 'input_values', key_type=str)
    return _execute_pipeline(
        pipeline,
        InMemoryInputManager(context, input_values, from_solids, through_solids),
        throw_on_error,
    )

def _execute_pipeline(
    pipeline,
    input_manager,
    throw_on_error=True,
):
    '''
    "Synchronous" version of execute_pipeline_iteator.

    throw_on_error makes the function throw when an error is encoutered rather than returning
    the SolidExecutionResult in an error-state.

    Note: throw_on_error is very useful in testing contexts when not testing for error conditions
    '''
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.inst_param(input_manager, 'input_manager', InputManager)
    check.bool_param(throw_on_error, 'throw_on_error')

    results = []
    with input_manager.yield_context() as context:
        for result in _execute_pipeline_iterator(
            context,
            pipeline,
            input_manager=input_manager,
        ):
            if throw_on_error:
                if not result.success:
                    _do_throw_on_error(result)

            results.append(result.copy())
        return DagsterPipelineExecutionResult(context, results)


class MaterializationArgs:
    def __init__(self, pipeline, materializations):
        check.inst_param(pipeline, 'pipeline', PipelineDefinition)
        check.list_param(materializations, 'materializations', of_type=config.Materialization)

        self.pipeline = pipeline
        self.materializations = list(materializations)
        self.through_solids = [materialization.solid for materialization in self.materializations]

    def should_materialize(self, solid_name):
        return solid_name in self.through_solids

    def materializations_for_solid(self, solid_name):
        for materialization in self.materializations:
            if materialization.solid == solid_name:
                yield materialization



import sqlite3
import sqlalchemy as sa


def test_raw_sqlite():
    conn = sqlite3.connect(':memory:')
    conn.cursor().execute('''CREATE TABLE num_table(num1 int, num2 int)''')
    conn.cursor().execute('''INSERT INTO num_table VALUES(1, 2)''')
    conn.cursor().execute('''INSERT INTO num_table VALUES(3, 4)''')
    conn.commit()

    fetch_cursor = conn.cursor()
    fetch_cursor.execute('SELECT * FROM num_table')
    results = fetch_cursor.fetchall()
    assert results == [(1, 2), (3, 4)]


def test_sqlalchemy_ddl():
    engine = sa.create_engine('sqlite://')
    create_num_table(engine)

    conn = engine.connect()
    results = conn.execute('SELECT * FROM num_table').fetchall()
    assert results == [(1, 2), (3, 4)]


def test_sqlalchemy_create_table_from_select():
    engine = sa.create_engine('sqlite://')

    create_num_table(engine)

    conn = engine.connect()

    conn.execute(
        '''CREATE TABLE sum_table AS SELECT num1, num2, num1 + num2 as sum FROM num_table'''
    )

    results = conn.execute('SELECT * FROM sum_table').fetchall()
    assert results == [(1, 2, 3), (3, 4, 7)]


def create_num_table(engine):
    metadata = sa.MetaData(engine)

    table = sa.Table(
        'num_table',
        metadata,
        sa.Column('num1', sa.Integer),
        sa.Column('num2', sa.Integer),
    )

    table.create()

    conn = engine.connect()

    conn.execute('''INSERT INTO num_table VALUES(1, 2)''')
    conn.execute('''INSERT INTO num_table VALUES(3, 4)''')



from __future__ import (absolute_import, division, print_function, unicode_literals)
from builtins import *  # pylint: disable=W0622,W0401
from .templated import execute_sql_text_on_context
from .common import (create_sql_alchemy_context_from_engine, SqlAlchemyResource)
from .subquery_builder_experimental import sql_file_solid



import sqlalchemy as sa
import dagster
from dagster import config
from dagster.core.execution import (
    output_single_solid,
    execute_pipeline,
)

from dagster.sqlalchemy_kernel.subquery_builder_experimental import (
    create_sql_solid,
    create_table_expression_input,
    create_table_input_dependency,
)
from .math_test_db import in_mem_context


def pipeline_test_def(solids, context):
    return dagster.PipelineDefinition(
        solids=solids,
        context_definitions={
            'default':
            dagster.PipelineContextDefinition(
                argument_def_dict={},
                context_fn=lambda _pipeline, _args: context,
            ),
        }
    )


def create_num_table(engine):
    metadata = sa.MetaData(engine)

    table = sa.Table(
        'num_table',
        metadata,
        sa.Column('num1', sa.Integer),
        sa.Column('num2', sa.Integer),
    )

    table.create()

    conn = engine.connect()

    conn.execute('''INSERT INTO num_table VALUES(1, 2)''')
    conn.execute('''INSERT INTO num_table VALUES(3, 4)''')


def test_sql_sum_solid():
    sum_table_solid = create_sum_table_solid()

    environment = config.Environment(
        sources={
            'sum_table': {
                'num_table': config.Source('TABLENAME', {'table_name': 'num_table'})
            }
        },
    )

    context = in_mem_context()

    result = output_single_solid(
        context,
        sum_table_solid,
        environment=environment,
        name='CREATE',
        arg_dict={'table_name': 'sum_table'},
    )
    assert result.success

    results = context.resources.sa.engine.connect().execute('SELECT * FROM sum_table').fetchall()
    assert results == [(1, 2, 3), (3, 4, 7)]


def create_sum_table_solid():
    return create_sql_solid(
        name='sum_table',
        inputs=[create_table_expression_input('num_table')],
        sql_text='SELECT num1, num2, num1 + num2 as sum FROM {num_table}',
    )


def create_sum_sq_pipeline(context):
    sum_solid = create_sum_table_solid()

    sum_sq_solid = create_sql_solid(
        name='sum_sq_table',
        inputs=[create_table_input_dependency(sum_solid)],
        sql_text='SELECT num1, num2, sum, sum * sum as sum_sq from {sum_table}',
    )

    return pipeline_test_def(solids=[sum_solid, sum_sq_solid], context=context)


def test_execute_sql_sum_sq_solid():
    pipeline = create_sum_sq_pipeline(in_mem_context())

    environment = config.Environment(
        sources={
            'sum_table': {
                'num_table': config.Source('TABLENAME', {'table_name': 'num_table'})
            }
        },
    )

    pipeline_result = execute_pipeline(
        pipeline,
        environment=environment,
    )

    assert pipeline_result.success

    result_list = pipeline_result.result_list

    sum_table_sql_text = result_list[0].transformed_value.query_text
    assert sum_table_sql_text == 'SELECT num1, num2, num1 + num2 as sum FROM num_table'

    sum_sq_table_sql_text = result_list[1].transformed_value.query_text
    assert sum_sq_table_sql_text == 'SELECT num1, num2, sum, sum * sum as sum_sq from ' + \
            '(SELECT num1, num2, num1 + num2 as sum FROM num_table)'


def test_output_sql_sum_sq_solid():
    pipeline = create_sum_sq_pipeline(in_mem_context())

    environment = config.Environment(
        sources={
            'sum_table': {
                'num_table': config.Source('TABLENAME', {'table_name': 'num_table'})
            }
        },
        materializations=[
            config.Materialization(
                solid='sum_sq_table',
                name='CREATE',
                args={'table_name': 'sum_sq_table'},
            )
        ],
    )

    pipeline_result = execute_pipeline(pipeline=pipeline, environment=environment)

    assert pipeline_result.success

    result_list = pipeline_result.result_list

    assert len(result_list) == 2
    engine = pipeline_result.context.resources.sa.engine
    result_list = engine.connect().execute('SELECT * FROM sum_sq_table').fetchall()
    assert result_list == [(1, 2, 3, 9), (3, 4, 7, 49)]



import copy
import json
import logging
from logging import (CRITICAL, DEBUG, ERROR, INFO, WARNING)
import traceback

import coloredlogs

from dagster import check

# This global, while fundamentally offensive, is necessary if I want to be able
# to apply a blanket logging level after the fact based on context
REGISTERED_LOGGERS = set()

class CompositeLogger:
    def __init__(self, loggers=None, level=logging.ERROR):
        self.loggers = check.opt_list_param(loggers, 'loggers', of_type=logging.Logger)

        for logger in self.loggers:
            logger.setLevel(level)

    def __getattr__(self, name):
        def _invoke_logger_method(*args, **kwargs):
            for logger in self.loggers:
                logger_method = check.is_callable(getattr(logger, name))
                logger_method(*args, **kwargs)
        return _invoke_logger_method

class JsonFileHandler(logging.Handler):
    def __init__(self, json_path):
        super().__init__()
        self.json_path = json_path

    def emit(self, record):
        try:
            logged_properties = copy.copy(record.__dict__)
            del logged_properties['msg']
            with open(self.json_path, 'a') as ff:
                text_line = json.dumps(logged_properties)
                ff.write(text_line + '\n')
        # Need to catch Exception here, so disabling lint
        except Exception as e: # pylint: disable=W0703
            logging.critical('Error during logging!')
            logging.exception(str(e))


def define_json_file_logger(name, json_path, level):
    check.str_param(name, 'name')

    if name in REGISTERED_LOGGERS:
        return logging.getLogger(name)

    logger = logging.getLogger(name)
    logger.setLevel(level)


    stream_handler = JsonFileHandler(json_path)
    stream_handler.setFormatter(define_default_formatter())
    logger.addHandler(stream_handler)

    return logger


def define_logger(name, level=INFO):
    check.str_param(name, 'name')

    if name in REGISTERED_LOGGERS:
        return logging.getLogger(name)

    logger = logging.getLogger(name)

    # This is the pre-coloredlogs process. Retaining here for posterity
    # stream_handler = logging.StreamHandler()
    # logger.setLevel(INFO)
    # stream_handler.setFormatter(define_default_formatter())
    # logger.addHandler(stream_handler)

    coloredlogs.install(logger=logger, level=level, fmt=default_format_string())

    REGISTERED_LOGGERS.add(name)

    return logger

def set_global_logging_level(level):
    check.param_invariant(
        level in {DEBUG, INFO, WARNING, ERROR, CRITICAL},
        'Invalid logging level {level}'.format(level=level)
    )
    for logger in REGISTERED_LOGGERS:
        logging.getLogger(logger).setLevel(level)

def add_global_handler(handler):
    handler.setFormatter(define_default_formatter())
    for logger in REGISTERED_LOGGERS:
        logging.getLogger(logger).addHandler(handler)

def default_format_string():
    return '%(asctime)s - %(name)s - %(levelname)s - %(message)s'

def define_default_formatter():
    return logging.Formatter(default_format_string())

def debug_format_string():
    return '''%(name)s.%(levelname)s: %(message)s
    time: %(asctime)s relative: %(relativeCreated)dms
    path: %(pathname)s line: %(lineno)d'''

def define_debug_formatter():
    return logging.Formatter(debug_format_string())


def get_formatted_stack_trace(exception):
    check.inst_param(exception, 'exception', Exception)
    return ''.join(traceback.format_tb(exception.__traceback__))



from __future__ import (absolute_import, division, print_function, unicode_literals)
from builtins import *  # pylint: disable=W0622,W0401
import inspect

from dagster import check


def has_context_argument(fn):
    check.callable_param(fn, 'fn')

    argspec = inspect.getfullargspec(fn)
    return 'context' in argspec[0]


def make_context_arg_optional(fn):
    check.callable_param(fn, 'fn')
    # make this idempotent during the transition
    if fn.__name__ == '__wrapper_with_context':
        return fn
    check.invariant(fn.__name__ != '__wrapper_with_context')

    if not has_context_argument(fn):

        def __wrapper_with_context(*args, context, **kwargs):
            check.not_none_param(context, 'context')
            return fn(*args, **kwargs)

        return __wrapper_with_context
    else:
        return fn



import jinja2

import dagster
from dagster import check
from dagster.core.definitions import (
    SolidDefinition,
    InputDefinition,
    SourceDefinition,
    ArgumentDefinition,
)

from .common import execute_sql_text_on_context


def _create_table_input(name, depends_on=None):
    return InputDefinition(
        name=name,
        sources=[
            SourceDefinition(
                source_type='TABLENAME',
                source_fn=lambda context, arg_dict: arg_dict,
                argument_def_dict={'name': ArgumentDefinition(dagster.core.types.String)},
            )
        ],
        depends_on=depends_on
    )


def create_templated_sql_transform_solid(
    name,
    sql,
    table_arguments,
    output,
    dependencies=None,
    extra_inputs=None,
):
    '''
    Create a solid that is a templated sql statement. This assumes that the sql statement
    is creating or modifying a table, and that that table will be used downstream in the pipeline
    for further transformations

    Usage example:

    sum_sql_template = 'CREATE TABLE {{sum_table.name}} AS ' + \
        'SELECT num1, num2, num1 + num2 as sum FROM {{num_table.name}}'

    sum_solid = create_templated_sql_transform_solid(
        name='sum_solid',
        sql=sum_sql_template,
        table_arguments=['sum_table', 'num_table'],
        output='sum_table',
    )

    'Table arguments' are inputs that represent a table. They have a string property "name". Each
    table argument ends up as an InputDefinition into the solid.

    To invoke this you would configure an argument dictionary to look like:

    input_arg_dict = {'sum_table': {'name': 'a_sum_table'}, 'num_table': {'name': 'a_num_table'}}

    dagster.execute_pipeline(context, pipeline, input_arg_dict)

    The output table is what is flowed to the *next* solid.

    So imagine, we built another solid which depended on the previous one.

    sum_sq_sql_template = 'CREATE TABLE {{sum_sq_table.name}} AS ' + \
        'SELECT num1, num2, sum, sum * sum as sum_sq FROM {{sum_solid.name}}'

    sum_sq_solid = create_templated_sql_transform_solid(
        name='sum_sq_solid',
        sql=sum_sq_sql_template,
        table_arguments=['sum_sq_table'],
        output='sum_sq_table',
        dependencies=[sum_solid],
    )

    This solid has two inputs, one is a table_argument directly passed in, and one is
    the table from the previous solid.

    Note: there is current awkwardness in this api because all inputs and solids in pipeline
    must have a different name. This is why the template variable in the "sum_sq_solid" is
    named "sum_solid" rather than "sum_table".

    To invoke this you would now specifiy:

    input_arg_dict = {
        'sum_table': {'name': 'a_sum_table'},
        'num_table': {'name': 'a_num_table'},
        'sum_sq_table': {'name': 'a_sum_sq_table'},
    }

    If you wanted to test the sum_sq_table in isolation, you could skip the num_table and just
    specify the output of the sum_solid:

    input_arg_dict = {
        'sum_solid': {'name': 'a_sum_table'},
        'sum_sq_table': {'name': 'a_sum_sq_table'},
    }

    As noted above the naming is not very intuitive right row, but the functionality works.
    '''
    check.str_param(name, 'name')
    check.str_param(sql, 'sql')
    check.list_param(table_arguments, 'table_arguments', of_type=str)
    check.str_param(output, 'output')
    dependencies = check.opt_list_param(dependencies, 'dependencies', of_type=SolidDefinition)
    extra_inputs = check.opt_list_param(extra_inputs, 'extra_inputs', of_type=InputDefinition)

    table_inputs = [_create_table_input(table) for table in table_arguments]
    dep_inputs = [_create_table_input(dep.name, depends_on=dep) for dep in dependencies]
    return SolidDefinition(
        name=name,
        inputs=table_inputs + dep_inputs + extra_inputs,
        transform_fn=_create_templated_sql_transform_with_output(sql, output),
        output=dagster.OutputDefinition(),
    )


def _render_template_string(template_text, args):
    template = jinja2.Environment(loader=jinja2.BaseLoader).from_string(template_text)
    return template.render(**args)


def _create_templated_sql_transform_with_output(sql, output_table):
    def do_transform(context, args):
        rendered_sql = _render_template_string(sql, args)
        execute_sql_text_on_context(context, rendered_sql)
        return args[output_table]

    return do_transform



import dagster
from dagster import config

from dagster.core.definitions import (
    SolidDefinition,
    InputDefinition,
    SourceDefinition,
    ArgumentDefinition,
)

from dagster.core.execution import (ExecutionContext, execute_single_solid, execute_pipeline)

from dagster.core import types


def test_execute_solid_with_input_same_name():
    solid = SolidDefinition(
        'a_thing',
        inputs=[
            InputDefinition(
                name='a_thing',
                sources=[
                    SourceDefinition(
                        source_type='a_source_type',
                        source_fn=lambda context, arg_dict: arg_dict['an_arg'],
                        argument_def_dict={'an_arg': ArgumentDefinition(types.String)},
                    ),
                ],
            ),
        ],
        transform_fn=lambda context, args: args['a_thing'] + args['a_thing'],
        output=dagster.OutputDefinition(),
    )

    result = execute_single_solid(
        ExecutionContext(),
        solid,
        environment=config.Environment(
            sources={
                'a_thing': {
                    'a_thing': config.Source(name='a_source_type', args={'an_arg': 'foo'})
                }
            }
        )
    )

    assert result.success
    assert result.transformed_value == 'foofoo'


def test_execute_two_solids_with_same_input_name():
    input_def = InputDefinition(
        name='a_thing',
        sources=[
            SourceDefinition(
                source_type='a_source_type',
                source_fn=lambda context, arg_dict: arg_dict['an_arg'],
                argument_def_dict={'an_arg': ArgumentDefinition(types.String)},
            ),
        ],
    )

    solid_one = SolidDefinition(
        'solid_one',
        inputs=[input_def],
        transform_fn=lambda context, args: args['a_thing'] + args['a_thing'],
        output=dagster.OutputDefinition(),
    )

    solid_two = SolidDefinition(
        'solid_two',
        inputs=[input_def],
        transform_fn=lambda context, args: args['a_thing'] + args['a_thing'],
        output=dagster.OutputDefinition(),
    )

    pipeline = dagster.PipelineDefinition(solids=[solid_one, solid_two])

    result = execute_pipeline(
        pipeline,
        environment=config.Environment(
            sources={
                'solid_one': {
                    'a_thing': config.Source(name='a_source_type', args={'an_arg': 'foo'})
                },
                'solid_two': {
                    'a_thing': config.Source(name='a_source_type', args={'an_arg': 'bar'})
                },
            }
        )
    )

    assert result.success

    assert result.result_named('solid_one').transformed_value == 'foofoo'
    assert result.result_named('solid_two').transformed_value == 'barbar'


def test_execute_dep_solid_different_input_name():
    first_solid = SolidDefinition(
        'first_solid',
        inputs=[
            InputDefinition(
                name='a_thing',
                sources=[
                    SourceDefinition(
                        source_type='a_source_type',
                        source_fn=lambda context, arg_dict: arg_dict['an_arg'],
                        argument_def_dict={'an_arg': ArgumentDefinition(types.String)},
                    ),
                ],
            ),
        ],
        transform_fn=lambda context, args: args['a_thing'] + args['a_thing'],
        output=dagster.OutputDefinition(),
    )

    second_solid = SolidDefinition(
        'second_solid',
        inputs=[
            InputDefinition(
                name='a_dependency',
                depends_on=first_solid,
            ),
        ],
        transform_fn=lambda context, args: args['a_dependency'] + args['a_dependency'],
        output=dagster.OutputDefinition(),
    )

    pipeline = dagster.PipelineDefinition(solids=[first_solid, second_solid])
    result = dagster.execute_pipeline(
        pipeline,
        environment=config.Environment(
            sources={
                'first_solid': {
                    'a_thing': config.Source(name='a_source_type', args={'an_arg': 'bar'})
                }
            }
        )
    )

    assert result.success
    assert len(result.result_list) == 2
    assert result.result_list[0].transformed_value == 'barbar'
    assert result.result_list[1].transformed_value == 'barbarbarbar'


def test_execute_dep_solid_same_input_name():
    def s_fn(arg_dict, executed, key):
        executed[key] = True
        return arg_dict

    executed = {
        's1_t1_source': False,
        's2_t1_source': False,
        's2_t2_source': False,
    }

    table_one = SolidDefinition(
        'table_one',
        inputs=[
            InputDefinition(
                name='table_one',
                sources=[
                    SourceDefinition(
                        source_type='TABLE',
                        source_fn=
                        lambda context, arg_dict: s_fn(arg_dict, executed, 's1_t1_source'),
                        argument_def_dict={'name': ArgumentDefinition(types.String)},
                    ),
                ],
            ),
        ],
        transform_fn=lambda context, args: args['table_one'],
        output=dagster.OutputDefinition(),
    )

    table_two = SolidDefinition(
        'table_two',
        inputs=[
            InputDefinition(
                name='table_one',
                sources=[
                    SourceDefinition(
                        source_type='TABLE',
                        source_fn=
                        lambda context, arg_dict: s_fn(arg_dict, executed, 's2_t1_source'),
                        argument_def_dict={'name': ArgumentDefinition(types.String)},
                    ),
                ],
                depends_on=table_one,
            ),
            InputDefinition(
                name='table_two',
                sources=[
                    SourceDefinition(
                        source_type='TABLE',
                        source_fn=
                        lambda context, arg_dict: s_fn(arg_dict, executed, 's2_t2_source'),
                        argument_def_dict={'name': ArgumentDefinition(types.String)},
                    ),
                ],
            ),
        ],
        transform_fn=lambda context, args: args['table_two'],
        output=dagster.OutputDefinition(),
    )

    pipeline = dagster.PipelineDefinition(solids=[table_one, table_two])

    sources = {
        'table_one': {
            'table_one': config.Source(name='TABLE', args={'name': 'table_one_instance'}),
        },
        'table_two': {
            'table_one': config.Source(name='TABLE', args={'name': 'table_one_instance'}),
            'table_two': config.Source(name='TABLE', args={'name': 'table_two_instance'}),
        },
    }

    complete_environment = config.Environment(sources=sources)

    both_solids_result = dagster.execute_pipeline(pipeline, environment=complete_environment)

    assert executed == {
        's1_t1_source': True,
        's2_t1_source': False,
        's2_t2_source': True,
    }

    assert both_solids_result.success

    assert len(both_solids_result.result_list) == 2
    assert both_solids_result.result_list[0].transformed_value == {'name': 'table_one_instance'}
    assert both_solids_result.result_list[1].transformed_value == {'name': 'table_two_instance'}

    # reset execution marks
    executed['s1_t1_source'] = False
    executed['s2_t1_source'] = False
    executed['s2_t2_source'] = False

    second_only_env = config.Environment(
        sources=sources,
        execution=config.Execution(from_solids=['table_two']),
    )

    second_solid_only_result = dagster.execute_pipeline(pipeline, environment=second_only_env)

    assert second_solid_only_result.success
    assert len(second_solid_only_result.result_list) == 1
    assert second_solid_only_result.result_list[0].name == 'table_two'
    assert second_solid_only_result.result_list[0].transformed_value == {
        'name': 'table_two_instance'
    }

    assert executed == {
        's1_t1_source': False,
        's2_t1_source': True,
        's2_t2_source': True,
    }

    # reset execution marks
    executed['s1_t1_source'] = False
    executed['s2_t1_source'] = False
    executed['s2_t2_source'] = False

    first_only_env = config.Environment(
        sources=sources,
        execution=config.Execution(through_solids=['table_one']),
    )

    first_solid_only_result = dagster.execute_pipeline(
        pipeline,
        environment=first_only_env,
    )

    assert first_solid_only_result.success
    assert len(first_solid_only_result.result_list) == 1
    assert first_solid_only_result.result_list[0].name == 'table_one'
    assert first_solid_only_result.result_list[0].transformed_value == {
        'name': 'table_one_instance'
    }

    assert executed == {
        's1_t1_source': True,
        's2_t1_source': False,
        's2_t2_source': False,
    }



import os
import sys

import click

from .context import Config
from .pipeline import create_pipeline_cli


def create_dagster_cli():
    group = click.group()(dagster_cli)
    group.add_command(create_pipeline_cli())
    return group


@click.option(
    '--config',
    '-c',
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
    ),
    default='pipelines.yml',
    help="Path to config file. Defaults to ./pipelines.yml."
)
@click.pass_context
def dagster_cli(ctx, config):
    ctx.obj = Config.from_file(config)
    sys.path.append(os.getcwd())



import graphviz

from dagster import check

from dagster.core.definitions import PipelineDefinition


def build_graphviz_graph(pipeline):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    graphviz_graph = graphviz.Graph('pipeline', directory='/tmp/graphviz')
    for solid in pipeline.solids:
        graphviz_graph.node(solid.name)

    graphviz_graph.attr('node', color='grey', shape='box')

    for solid in pipeline.solids:
        for input_def in solid.inputs:
            scoped_name = solid.name + '.' + input_def.name
            graphviz_graph.node(scoped_name)

    for solid in pipeline.solids:
        for input_def in solid.inputs:
            scoped_name = solid.name + '.' + input_def.name
            graphviz_graph.edge(scoped_name, solid.name)

            if input_def.depends_on:
                graphviz_graph.edge(input_def.depends_on.name, scoped_name)

    return graphviz_graph



import dagster
from dagster import (config, check)
from dagster.core.definitions import (
    InputDefinition,
    SolidDefinition,
    SourceDefinition,
    ArgumentDefinition,
)
from dagster.sqlalchemy_kernel.templated import (
    _create_templated_sql_transform_with_output,
    _render_template_string,
    create_templated_sql_transform_solid,
)

from .math_test_db import in_mem_context


def _load_table(context, table_name):
    return context.resources.sa.engine.connect().execute(f'SELECT * FROM {table_name}').fetchall()


def table_name_source(table_name):
    check.str_param(table_name, 'table_name')
    return config.Source('TABLENAME', args={'name': table_name})


def pipeline_test_def(solids, context):
    return dagster.PipelineDefinition(
        solids=solids,
        context_definitions={
            'default':
            dagster.PipelineContextDefinition(
                argument_def_dict={},
                context_fn=lambda _pipeline, _args: context,
            ),
        }
    )


def test_single_templated_sql_solid_single_table_raw_api():
    sql = '''CREATE TABLE {{sum_table.name}}
    AS SELECT num1, num2, num1 + num2 as sum FROM num_table'''

    sum_table_arg = 'specific_sum_table'

    sum_table_input = InputDefinition(
        name='sum_table',
        sources=[
            SourceDefinition(
                source_type='TABLENAME',
                source_fn=lambda context, arg_dict: arg_dict,
                argument_def_dict={
                    'name': ArgumentDefinition(dagster.core.types.String),
                },
            )
        ]
    )

    sum_table_transform_solid = SolidDefinition(
        name='sum_table_transform',
        inputs=[sum_table_input],
        transform_fn=_create_templated_sql_transform_with_output(sql, 'sum_table'),
        output=dagster.OutputDefinition(),
    )

    pipeline = pipeline_test_def(solids=[sum_table_transform_solid], context=in_mem_context())
    environment = config.Environment(
        sources={'sum_table_transform': {
            'sum_table': table_name_source(sum_table_arg)
        }},
    )

    result = dagster.execute_pipeline(pipeline, environment=environment)
    assert result.success

    assert _load_table(result.context, sum_table_arg) == [(1, 2, 3), (3, 4, 7)]

    environment_without_source = config.Environment(
        sources={'sum_table_transform': {
            'sum_table': table_name_source('another_table')
        }},
    )
    result_no_source = dagster.execute_pipeline(pipeline, environment=environment_without_source)
    assert result_no_source.success

    assert _load_table(result_no_source.context, 'another_table') == [(1, 2, 3), (3, 4, 7)]


def test_single_templated_sql_solid_single_table_with_api():

    sql = '''CREATE TABLE {{sum_table.name}} AS
    SELECT num1, num2, num1 + num2 as sum FROM num_table'''

    sum_table_arg = 'specific_sum_table'

    sum_table_transform = create_templated_sql_transform_solid(
        name='sum_table_transform',
        sql=sql,
        table_arguments=['sum_table'],
        output='sum_table',
    )

    pipeline = pipeline_test_def(solids=[sum_table_transform], context=in_mem_context())

    environment = config.Environment(
        sources={'sum_table_transform': {
            'sum_table': table_name_source(sum_table_arg)
        }}
    )

    result = dagster.execute_pipeline(pipeline, environment=environment)
    assert result.success

    assert _load_table(result.context, sum_table_arg) == [(1, 2, 3), (3, 4, 7)]


def test_single_templated_sql_solid_double_table_raw_api():
    sum_table_arg = 'specific_sum_table'
    num_table_arg = 'specific_num_table'

    sql = '''CREATE TABLE {{sum_table.name}} AS
        SELECT num1, num2, num1 + num2 as sum FROM {{num_table.name}}'''

    sum_table_input = InputDefinition(
        name='sum_table',
        sources=[
            SourceDefinition(
                source_type='TABLENAME',
                source_fn=lambda context, arg_dict: arg_dict,
                argument_def_dict={
                    'name': ArgumentDefinition(dagster.core.types.String),
                },
            )
        ]
    )

    num_table_input = InputDefinition(
        name='num_table',
        sources=[
            SourceDefinition(
                source_type='TABLENAME',
                source_fn=lambda context, arg_dict: arg_dict,
                argument_def_dict={
                    'name': ArgumentDefinition(dagster.core.types.String),
                },
            )
        ]
    )

    sum_solid = SolidDefinition(
        name='sum_solid',
        inputs=[sum_table_input, num_table_input],
        transform_fn=_create_templated_sql_transform_with_output(
            sql,
            'sum_table',
        ),
        output=dagster.OutputDefinition(),
    )

    pipeline = pipeline_test_def(solids=[sum_solid], context=in_mem_context(num_table_arg))

    environment = config.Environment(
        sources={
            'sum_solid': {
                'sum_table': table_name_source(sum_table_arg),
                'num_table': table_name_source(num_table_arg),
            },
        },
    )

    result = dagster.execute_pipeline(pipeline, environment=environment)
    assert result.success

    assert _load_table(result.context, sum_table_arg) == [(1, 2, 3), (3, 4, 7)]


def test_single_templated_sql_solid_double_table_with_api():
    sum_table_arg = 'specific_sum_table'
    num_table_arg = 'specific_num_table'

    sql = '''CREATE TABLE {{sum_table.name}} AS
    SELECT num1, num2, num1 + num2 as sum FROM {{num_table.name}}'''

    sum_solid = create_templated_sql_transform_solid(
        name='sum_solid',
        sql=sql,
        table_arguments=['sum_table', 'num_table'],
        output='sum_table',
    )

    pipeline = pipeline_test_def(solids=[sum_solid], context=in_mem_context(num_table_arg))

    environment = config.Environment(
        sources={
            'sum_solid': {
                'sum_table': table_name_source(sum_table_arg),
                'num_table': table_name_source(num_table_arg),
            },
        },
    )

    result = dagster.execute_pipeline(pipeline, environment=environment)
    assert result.success

    assert _load_table(result.context, sum_table_arg) == [(1, 2, 3), (3, 4, 7)]


def test_templated_sql_solid_pipeline():
    sum_sql_template = '''CREATE TABLE {{sum_table.name}} AS
        SELECT num1, num2, num1 + num2 as sum FROM num_table'''

    sum_sq_sql_template = '''CREATE TABLE {{sum_sq_table.name}} AS
        SELECT num1, num2, sum, sum * sum as sum_sq FROM {{sum_table.name}}'''

    sum_solid = create_templated_sql_transform_solid(
        name='sum_table',
        sql=sum_sql_template,
        table_arguments=['sum_table'],
        output='sum_table',
    )

    sum_sq_solid = create_templated_sql_transform_solid(
        name='sum_sq_table',
        sql=sum_sq_sql_template,
        table_arguments=['sum_sq_table'],
        output='sum_sq_table',
        dependencies=[sum_solid],
    )

    context = in_mem_context()

    pipeline = pipeline_test_def(solids=[sum_solid, sum_sq_solid], context=context)
    first_sum_table = 'first_sum_table'
    first_sum_sq_table = 'first_sum_sq_table'

    environment_one = config.Environment(
        sources={
            'sum_table': {
                'sum_table': table_name_source(first_sum_table)
            },
            'sum_sq_table': {
                'sum_sq_table': table_name_source(first_sum_sq_table)
            },
        }
    )
    first_result = dagster.execute_pipeline(pipeline, environment=environment_one)
    assert first_result.success

    assert len(first_result.result_list) == 2
    assert first_result.result_list[0].transformed_value == {'name': first_sum_table}
    assert first_result.result_list[1].transformed_value == {'name': first_sum_sq_table}

    assert _load_table(first_result.context, first_sum_table) == [(1, 2, 3), (3, 4, 7)]

    assert _load_table(first_result.context, first_sum_sq_table) == [(1, 2, 3, 9), (3, 4, 7, 49)]

    # now execute subdag

    pipeline_two = pipeline_test_def(solids=[sum_solid, sum_sq_solid], context=context)

    second_sum_sq_table = 'second_sum_sq_table'

    environment_two = config.Environment(
        sources={
            'sum_sq_table': {
                'sum_table': table_name_source(first_sum_table),
                'sum_sq_table': table_name_source(second_sum_sq_table),
            },
        },
        execution=config.Execution.single_solid('sum_sq_table'),
    )

    second_result = dagster.execute_pipeline(
        pipeline_two,
        environment=environment_two,
    )
    assert second_result.success
    assert len(second_result.result_list) == 1
    assert _load_table(second_result.context, second_sum_sq_table) == [(1, 2, 3, 9), (3, 4, 7, 49)]


def test_templated_sql_solid_with_api():
    sql_template = '''CREATE TABLE {{sum_table.name}} AS
        SELECT num1, num2, num1 + num2 as sum FROM num_table'''

    sum_solid = create_templated_sql_transform_solid(
        name='sum_solid',
        sql=sql_template,
        table_arguments=['sum_table'],
        output='sum_table',
    )

    pipeline = pipeline_test_def(solids=[sum_solid], context=in_mem_context())

    sum_table_arg = 'specific_sum_table'
    environment = config.Environment(
        sources={'sum_solid': {
            'sum_table': table_name_source(sum_table_arg)
        }}
    )
    result = dagster.execute_pipeline(pipeline, environment=environment)
    assert result.success

    assert _load_table(result.context, sum_table_arg) == [(1, 2, 3), (3, 4, 7)]


def test_with_from_through_specifying_all_solids():
    pipeline = create_multi_input_pipeline()

    first_sum_table = 'first_sum_table'
    first_mult_table = 'first_mult_table'
    first_sum_mult_table = 'first_sum_mult_table'

    all_solid_names = [solid.name for solid in pipeline.solids]

    environment = config.Environment(
        sources={
            'sum_table': {
                'sum_table': table_name_source(first_sum_table),
            },
            'mult_table': {
                'mult_table': table_name_source(first_mult_table),
            },
            'sum_mult_table': {
                'sum_mult_table': table_name_source(first_sum_mult_table),
            },
        },
        execution=config.Execution(from_solids=all_solid_names, through_solids=all_solid_names)
    )

    pipeline_result = dagster.execute_pipeline(pipeline, environment=environment)
    assert len(pipeline_result.result_list) == 3
    assert _load_table(pipeline_result.context, first_sum_table) == [(1, 2, 3), (3, 4, 7)]
    assert _load_table(pipeline_result.context, first_mult_table) == [(1, 2, 2), (3, 4, 12)]
    assert _load_table(pipeline_result.context, first_sum_mult_table) == [(1, 3, 2), (3, 7, 12)]


def test_multi_input_partial_execution():
    pipeline = create_multi_input_pipeline()

    first_sum_table = 'first_sum_table'
    first_mult_table = 'first_mult_table'
    first_sum_mult_table = 'first_sum_mult_table'

    environment = config.Environment(
        sources={
            'sum_table': {
                'sum_table': table_name_source(first_sum_table),
            },
            'mult_table': {
                'mult_table': table_name_source(first_mult_table),
            },
            'sum_mult_table': {
                'sum_mult_table': table_name_source(first_sum_mult_table),
            },
        }
    )

    first_pipeline_result = dagster.execute_pipeline(pipeline, environment=environment)

    assert first_pipeline_result.success
    assert len(first_pipeline_result.result_list) == 3
    assert _load_table(first_pipeline_result.context, first_sum_table) == [(1, 2, 3), (3, 4, 7)]
    assert _load_table(first_pipeline_result.context, first_mult_table) == [(1, 2, 2), (3, 4, 12)]
    assert _load_table(first_pipeline_result.context,
                       first_sum_mult_table) == [(1, 3, 2), (3, 7, 12)]

    second_sum_mult_table = 'second_sum_mult_table'

    environment_two = config.Environment(
        sources={
            'sum_mult_table': {
                'sum_table': table_name_source(first_sum_table),
                'mult_table': table_name_source(first_mult_table),
                'sum_mult_table': table_name_source(second_sum_mult_table)
            },
        },
        execution=config.Execution.single_solid('sum_mult_table')
    )

    second_pipeline_result = dagster.execute_pipeline(
        pipeline,
        environment=environment_two,
    )

    assert second_pipeline_result.success
    assert len(second_pipeline_result.result_list) == 1
    assert _load_table(second_pipeline_result.context,
                       second_sum_mult_table) == [(1, 3, 2), (3, 7, 12)]


def create_multi_input_pipeline():
    sum_sql_template = '''CREATE TABLE {{sum_table.name}} AS
        SELECT num1, num2, num1 + num2 as sum FROM num_table'''

    mult_sql_template = '''CREATE TABLE {{mult_table.name}} AS
        SELECT num1, num2, num1 * num2 as mult FROM num_table'''

    sum_mult_join_template = '''CREATE TABLE {{sum_mult_table.name}} AS
        SELECT {{sum_table.name}}.num1, sum, mult FROM {{sum_table.name}}
        INNER JOIN {{mult_table.name}} ON {{sum_table.name}}.num1 = {{mult_table.name}}.num1'''

    sum_solid = create_templated_sql_transform_solid(
        name='sum_table',
        sql=sum_sql_template,
        table_arguments=['sum_table'],
        output='sum_table',
    )

    mult_solid = create_templated_sql_transform_solid(
        name='mult_table',
        sql=mult_sql_template,
        table_arguments=['mult_table'],
        output='mult_table',
    )

    sum_mult_solid = create_templated_sql_transform_solid(
        name='sum_mult_table',
        sql=sum_mult_join_template,
        table_arguments=['sum_mult_table'],
        dependencies=[sum_solid, mult_solid],
        output='sum_mult_table',
    )

    pipeline = pipeline_test_def(
        solids=[sum_solid, mult_solid, sum_mult_solid], context=in_mem_context()
    )
    return pipeline


def test_jinja():
    templated_sql = '''SELECT * FROM {{some_table.name}}'''

    sql = _render_template_string(templated_sql, args={'some_table': {'name': 'fill_me_in'}})

    assert sql == '''SELECT * FROM fill_me_in'''



import dagster
from dagster import (solid, InputDefinition, OutputDefinition, PipelineDefinition)
import dagster.pandas_kernel as dagster_pd


@solid(
    inputs=[InputDefinition('num', dagster_pd.DataFrame)],
    output=OutputDefinition(dagster_pd.DataFrame)
)
def sum_solid(num):
    sum_df = num.copy()
    sum_df['sum'] = sum_df['num1'] + sum_df['num2']
    return sum_df


@solid(
    inputs=[InputDefinition('sum_df', dagster_pd.DataFrame, depends_on=sum_solid)],
    output=OutputDefinition(dagster_pd.DataFrame)
)
def sum_sq_solid(sum_df):
    sum_sq_df = sum_df.copy()
    sum_sq_df['sum_sq'] = sum_df['sum']**2
    return sum_sq_df


@solid(
    inputs=[InputDefinition('sum_sq_solid', dagster_pd.DataFrame, depends_on=sum_sq_solid)],
    output=OutputDefinition(dagster_pd.DataFrame)
)
def always_fails_solid(**_kwargs):
    raise Exception('I am a programmer and I make error')


def define_pipeline():
    return dagster.PipelineDefinition(
        name='pandas_hello_world_fails',
        solids=[
            sum_solid,
            sum_sq_solid,
            always_fails_solid,
        ],
    )


def define_success_pipeline():
    return PipelineDefinition(name='pandas_hello_world', solids=[sum_solid, sum_sq_solid])



from toposort import toposort_flatten

import dagster

from dagster import check
from dagster.core import types
from dagster.utils import logging

import dagster.core.definitions

# from .definitions import (SolidDefinition, PipelineContextDefinition)


def create_adjacency_lists(solids):
    check.list_param(solids, 'solids', of_type=dagster.core.definitions.SolidDefinition)

    visit_dict = {s.name: False for s in solids}
    forward_edges = {s.name: set() for s in solids}
    backward_edges = {s.name: set() for s in solids}

    def visit(solid):
        if visit_dict[solid.name]:
            return

        visit_dict[solid.name] = True

        for inp in solid.inputs:
            if inp.depends_on is not None:
                from_node = inp.depends_on.name
                to_node = solid.name
                if from_node in forward_edges:
                    forward_edges[from_node].add(to_node)
                    backward_edges[to_node].add(from_node)
                    visit(inp.depends_on)

    for s in solids:
        visit(s)

    return (forward_edges, backward_edges)


class SolidGraph:
    def __init__(self, solids):
        check.list_param(solids, 'solids', of_type=dagster.core.definitions.SolidDefinition)
        self._solid_dict = {solid.name: solid for solid in solids}

        solid_names = set([solid.name for solid in solids])
        check.invariant(len(solid_names) == len(solids), 'must have unique names')

        all_inputs = {}

        for solid in solids:
            for input_def in solid.inputs:
                # if input exists should probably ensure that it is the same
                all_inputs[input_def.name] = input_def

        self._all_inputs = all_inputs
        self.forward_edges, self.backward_edges = create_adjacency_lists(solids)
        self.topological_order = toposort_flatten(self.backward_edges, sort=True)

        self._transitive_deps = {}

    @property
    def topological_solids(self):
        return [self._solid_dict[name] for name in self.topological_order]

    @property
    def solids(self):
        return list(self._solid_dict.values())

    def transitive_dependencies_of(self, solid_name):
        check.str_param(solid_name, 'solid_name')

        if solid_name in self._transitive_deps:
            return self._transitive_deps[solid_name]

        trans_deps = set()
        for inp in self._solid_dict[solid_name].inputs:
            if inp.depends_on:
                trans_deps.add(inp.depends_on.name)
                trans_deps.union(self.transitive_dependencies_of(inp.depends_on.name))

        self._transitive_deps[solid_name] = trans_deps
        return self._transitive_deps[solid_name]

    def _check_solid_name(self, solid_name):
        check.str_param(solid_name, 'output_name')
        check.param_invariant(
            solid_name in self._solid_dict, 'output_name',
            f'Solid {solid_name} must exist in {list(self._solid_dict.keys())}'
        )

    def _check_input_names(self, input_names):
        check.list_param(input_names, 'input_names', of_type=str)

        for input_name in input_names:
            check.param_invariant(
                input_name in self._all_inputs, 'input_names', 'Input name not found'
            )

    def compute_unprovided_inputs(self, solid_name, input_names):
        '''
        Given a single solid_name and a set of input_names that represent the
        set of inputs provided for a computation, return the inputs that are *missing*.
        This detects the case where an upstream path in the DAG of solids does not
        have enough information to materialize a given output.
        '''

        self._check_solid_name(solid_name)
        self._check_input_names(input_names)

        input_set = set(input_names)

        unprovided_inputs = set()

        visit_dict = {name: False for name in self._solid_dict.keys()}

        output_solid = self._solid_dict[solid_name]

        def visit(solid):
            if visit_dict[solid.name]:
                return
            visit_dict[solid.name] = True

            for inp in solid.inputs:
                if inp.name in input_set:
                    continue

                if inp.is_external:
                    unprovided_inputs.add(inp.name)
                else:
                    visit(inp.depends_on)

        visit(output_solid)

        return unprovided_inputs

    def create_execution_subgraph(self, from_solids, to_solids):
        check.list_param(from_solids, 'from_solids', of_type=str)
        check.list_param(to_solids, 'to_solids', of_type=str)

        from_solid_set = set(from_solids)
        involved_solids = from_solid_set

        def visit(solid):
            if solid.name in involved_solids:
                return
            involved_solids.add(solid.name)

            for input_def in solid.inputs:
                if input_def.is_external:
                    continue

                if input_def.depends_on.name in from_solid_set:
                    continue

                visit(input_def.depends_on)

        for to_solid in to_solids:
            visit(self._solid_dict[to_solid])

        return SolidGraph([self._solid_dict[name] for name in involved_solids])



import pytest

from dagster import (ArgumentDefinition, types)
from dagster.core.argument_handling import validate_args
from dagster.core.errors import DagsterTypeError


def _validate(argument_def_dict, arg_dict):
    return validate_args(argument_def_dict, arg_dict, 'dummy')


def _single_required_string_arg_def_dict():
    return {'string_arg': ArgumentDefinition(types.String)}


def _multiple_required_args_def_dict():
    return {
        'arg_one': ArgumentDefinition(types.String),
        'arg_two': ArgumentDefinition(types.String)
    }


def _single_optional_string_arg_def_dict():
    return {'optional_arg': ArgumentDefinition(types.String, is_optional=True)}


def _single_optional_string_arg_def_dict_with_default():
    return {
        'optional_arg':
        ArgumentDefinition(types.String, is_optional=True, default_value='some_default')
    }


def _mixed_required_optional_string_arg_def_dict_with_default():
    return {
        'optional_arg':
        ArgumentDefinition(types.String, is_optional=True, default_value='some_default'),
        'required_arg':
        ArgumentDefinition(types.String, is_optional=False),
        'optional_arg_no_default':
        ArgumentDefinition(types.String, is_optional=True),
    }


def test_empty():
    _validate({}, {})


def test_required_single_arg_passing():
    _validate(_single_required_string_arg_def_dict(), {'string_arg': 'value'})
    _validate(_single_required_string_arg_def_dict(), {'string_arg': None})


def test_multiple_required_arg_passing():
    _validate(_multiple_required_args_def_dict(), {'arg_one': 'value_one', 'arg_two': 'value_two'})
    _validate(_multiple_required_args_def_dict(), {'arg_one': 'value_one', 'arg_two': None})


def test_single_required_arg_failures():
    with pytest.raises(DagsterTypeError):
        _validate(_single_required_string_arg_def_dict(), {})

    with pytest.raises(DagsterTypeError):
        _validate(_single_required_string_arg_def_dict(), {'extra': 'yup'})

    with pytest.raises(DagsterTypeError):
        _validate(_single_required_string_arg_def_dict(), {'string_arg': 'yupup', 'extra': 'yup'})

    with pytest.raises(DagsterTypeError):
        _validate(_single_required_string_arg_def_dict(), {'string_arg': 1})


def test_multiple_required_args_failing():
    with pytest.raises(DagsterTypeError):
        _validate(_multiple_required_args_def_dict(), {})

    with pytest.raises(DagsterTypeError):
        _validate(_multiple_required_args_def_dict(), {'arg_one': 'yup'})

    with pytest.raises(DagsterTypeError):
        _validate(_multiple_required_args_def_dict(), {'arg_one': 'yup', 'extra': 'yup'})

    with pytest.raises(DagsterTypeError):
        _validate(_multiple_required_args_def_dict(), {'arg_one': 'value_one', 'arg_two': 2})


def test_single_optional_arg_passing():
    assert _validate(_single_optional_string_arg_def_dict(), {'optional_arg': 'value'}) == {
        'optional_arg': 'value'
    }
    assert _validate(_single_optional_string_arg_def_dict(), {}) == {}

    assert _validate(_single_optional_string_arg_def_dict(), {'optional_arg': None}) == {
        'optional_arg': None
    }


def test_single_optional_arg_failing():
    with pytest.raises(DagsterTypeError):
        _validate(_single_optional_string_arg_def_dict(), {'optional_arg': 1})

    with pytest.raises(DagsterTypeError):
        _validate(_single_optional_string_arg_def_dict(), {'optional_argdfd': 1})


def test_single_optional_arg_passing_with_default():
    assert _validate(_single_optional_string_arg_def_dict_with_default(), {}) == {
        'optional_arg': 'some_default'
    }

    assert _validate(
        _single_optional_string_arg_def_dict_with_default(), {'optional_arg': 'override'}
    ) == {
        'optional_arg': 'override'
    }


def test_mixed_args_passing():
    assert _validate(
        _mixed_required_optional_string_arg_def_dict_with_default(), {
            'optional_arg': 'value_one',
            'required_arg': 'value_two',
        }
    ) == {
        'optional_arg': 'value_one',
        'required_arg': 'value_two',
    }

    assert _validate(
        _mixed_required_optional_string_arg_def_dict_with_default(), {
            'required_arg': 'value_two',
        }
    ) == {
        'optional_arg': 'some_default',
        'required_arg': 'value_two',
    }

    assert _validate(
        _mixed_required_optional_string_arg_def_dict_with_default(), {
            'required_arg': 'value_two',
            'optional_arg_no_default': 'value_three',
        }
    ) == {
        'optional_arg': 'some_default',
        'required_arg': 'value_two',
        'optional_arg_no_default': 'value_three',
    }



import copy
import keyword
import re

from dagster import check
from dagster.core import types

from .errors import DagsterInvalidDefinitionError
from .graph import SolidGraph

DISALLOWED_NAMES = set(
    [
        'context',
        'meta',
        'arg_dict',
        'dict',
        'input_arg_dict',
        'output_arg_dict',
        'int',
        'str',
        'float',
        'bool',
        'input',
        'output',
        'result',
        'type',
    ] + keyword.kwlist  # just disallow all python keywords
)


def check_valid_name(name):
    check.str_param(name, 'name')
    if name in DISALLOWED_NAMES:
        raise DagsterInvalidDefinitionError('{name} is not allowed'.format(name=name))

    regex = r'^[A-Za-z0-9_]+$'
    if not re.match(regex, name):
        raise DagsterInvalidDefinitionError(
            '{name} must be in regex {regex}'.format(name=name, regex=regex)
        )
    return name


def check_argument_def_dict(argument_def_dict):
    return check.dict_param(
        argument_def_dict,
        'argument_def_dict',
        key_type=str,
        value_type=ArgumentDefinition,
    )


class PipelineContextDefinition:
    def __init__(self, *, argument_def_dict, context_fn, description=None):
        self.argument_def_dict = check_argument_def_dict(argument_def_dict)
        self.context_fn = check.callable_param(context_fn, 'context_fn')
        self.description = description


def _default_pipeline_context_definitions():
    def _default_context_fn(_pipeline, args):
        # This has a circular dependency between execution and definition
        # The likely solution is to move the ExecutionContext to definitions.py
        # -- schrockn (07-28-18)
        import dagster.core.execution
        import dagster.utils.logging

        log_level = args['log_level']
        context = dagster.core.execution.ExecutionContext(
            log_level=log_level,
            loggers=[dagster.utils.logging.define_logger('dagster', level=log_level)]
        )
        return context

    default_context_def = PipelineContextDefinition(
        argument_def_dict={
            'log_level':
            ArgumentDefinition(dagster_type=types.String, is_optional=True, default_value='ERROR')
        },
        context_fn=_default_context_fn,
    )
    return {'default': default_context_def}


class PipelineDefinition:
    def __init__(self, solids, name=None, description=None, context_definitions=None):
        self.description = check.opt_str_param(description, 'description')
        self.name = check.opt_str_param(name, 'name')

        if context_definitions is None:
            context_definitions = _default_pipeline_context_definitions()

        self.context_definitions = check.dict_param(
            context_definitions,
            'context_definitions',
            key_type=str,
            value_type=PipelineContextDefinition,
        )

        for solid in solids:
            if not isinstance(solid, SolidDefinition) and callable(solid):
                raise DagsterInvalidDefinitionError(
                    '''You have passed a lambda or function {solid} into
                a pipeline that is not a solid. You have likely forgetten to annotate this function
                with an @solid decorator located in dagster.core.decorators
                '''.format(solid=solid.__name__)
                )

        self.solids = check.list_param(solids, 'solids', of_type=SolidDefinition)

        solid_names = set([solid.name for solid in self.solids])
        for solid in solids:
            for input_def in solid.inputs:
                if input_def.depends_on:
                    check.invariant(
                        input_def.depends_on.name in solid_names,
                        f'dep must exist got: {input_def.depends_on.name} and set {solid_names}'
                    )

        self.solid_graph = SolidGraph(solids=solids)

    @property
    def solid_names(self):
        return [solid.name for solid in self.solids]

    @property
    def input_names(self):
        return set([input_def.name for input_def in self.all_inputs])

    def get_input(self, solid_name, input_name):
        for solid in self.solids:
            if solid.name != solid_name:
                continue
            for input_def in solid.inputs:
                if input_def.name == input_name:
                    return input_def
        check.failed('not found')

    @property
    def external_inputs(self):
        for input_def in self.all_inputs:
            if input_def.is_external:
                yield input_def

    @property
    def externally_sourced_solids(self):
        for solid in self.solids:
            for input_def in solid.inputs:
                if input_def.is_external:
                    yield solid
                    break

    def has_solid(self, name):
        check.str_param(name, 'name')
        for solid in self.solids:
            if solid.name == name:
                return True
        return False

    @property
    def all_inputs(self):
        for solid in self.solids:
            for input_def in solid.inputs:
                yield input_def

    def solid_named(self, name):
        check.str_param(name, 'name')
        for solid in self.solids:
            if solid.name == name:
                return solid
        check.failed('Could not find solid named ' + name)

    @property
    def all_depended_on_solids(self):
        for input_def in self.all_inputs:
            if input_def.depends_on:
                yield input_def.depends_on

    @property
    def all_sink_solids(self):
        all_names = set([solid.name for solid in self.solids])
        all_depended_on_names = set([solid.name for solid in self.all_depended_on_solids])
        return all_names.difference(all_depended_on_names)


class ExpectationResult:
    def __init__(self, success, solid=None, message=None, result_context=None):
        self.success = check.bool_param(success, 'success')
        self.solid = check.opt_inst_param(solid, SolidDefinition, 'solid')
        self.message = check.opt_str_param(message, 'message')
        self.result_context = check.opt_dict_param(result_context, 'result_context')

    def copy(self):
        return copy.deepcopy(self)


class ExpectationDefinition:
    def __init__(self, name, expectation_fn, description=None):
        self.name = check_valid_name(name)
        self.expectation_fn = check.callable_param(expectation_fn, 'expectation_fn')
        self.description = check.opt_str_param(description, 'description')


class SourceDefinition:
    '''
    name: name of the source

    source_fn: callable
         The input function defines exactly what happens when the source is invoked.

         def simplified_read_csv_example(context, arg_dict):
             context.info('I am in an input.') # use context for logging
             return pd.read_csv(arg_dict['path'])

    argument_def_dict: { str: DagsterType }
         Define the arguments expected by this source . A dictionary that maps a string
         (argument name) to an argument type (defined in dagster.core.types) Continuing
         the above example, the csv signature would be:

         argument_def_dict = {'path' : dagster.core.types.Path }

    '''

    def __init__(self, source_type, source_fn, argument_def_dict, description=None):
        check.callable_param(source_fn, 'source_fn')
        self.source_type = check_valid_name(source_type)
        self.source_fn = check.callable_param(source_fn, 'source_fn')
        self.argument_def_dict = check_argument_def_dict(argument_def_dict)
        self.description = check.opt_str_param(description, 'description')


class InputDefinition:
    '''
    An InputDefinition instances represents an argument to a transform defined within a solid.

    - name: Name of input

    - sources: A list of possible sources for the input. For example, an input which is passed
    to the transform as a pandas dataframe could have any number of different source types
    (CSV, Parquet etc). Some inputs have zero sources, and can only be created by
    execute a dependant solid.

    - depends_on: (Optional). This input depends on another solid in the context of a
    a pipeline.

    - input_callback: (Optional) Called on the source result. Gets execution context and result.
    Can be used to validate the result, log stats etc.
    '''

    def __init__(
        self,
        name,
        dagster_type=None,
        sources=None,
        depends_on=None,
        expectations=None,
        input_callback=None,
        description=None
    ):
        self.name = check_valid_name(name)

        if sources is None and dagster_type is not None:
            sources = dagster_type.default_sources

        self.sources = check.opt_list_param(sources, 'sources', of_type=SourceDefinition)
        self.depends_on = check.opt_inst_param(depends_on, 'depends_on', SolidDefinition)

        self.dagster_type = check.opt_inst_param(
            dagster_type, 'dagster_type', types.DagsterType, types.Any
        )

        self.expectations = check.opt_list_param(
            expectations, 'expectations', of_type=ExpectationDefinition
        )
        self.input_callback = check.opt_callable_param(input_callback, 'input_callback')
        self.description = check.opt_str_param(description, 'description')

    @property
    def is_external(self):
        return self.depends_on is None

    def source_of_type(self, source_type):
        check.str_param(source_type, 'source_type')
        for source in self.sources:
            if source.source_type == source_type:
                return source

        check.failed(
            'Source {source_type} not found in input {input_name}.'.format(
                source_type=source_type,
                input_name=self.name,
            )
        )


class MaterializationDefinition:
    '''
    materialization_fn: callable

        This function defines the actual output.

        The first function argument is the result of the transform function. It can be
        named anything.

        You must specify an argument with the name "arg_dict". It will be the dictionary
        of arguments specified by the caller of the solid

        def materialization_fn(context, args, value):
            pass

    argument_def_dict: { str: DagsterType }
        Define the arguments expected by this materialization. A dictionary that maps a string
        (argument name) to an argument type (defined in dagster.core.types).

        e.g.:

        argument_def_dict = { 'path' : dagster.core.types.Path }
    '''

    def __init__(self, name, materialization_fn, argument_def_dict=None, description=None):
        self.name = check_valid_name(name)
        self.materialization_fn = check.callable_param(materialization_fn, 'materialization_fn')
        self.argument_def_dict = check_argument_def_dict(argument_def_dict)
        self.description = check.opt_str_param(description, 'description')


class OutputDefinition:
    # runtime type info
    def __init__(
        self,
        dagster_type=None,
        materializations=None,
        expectations=None,
        output_callback=None,
        description=None
    ):

        self.dagster_type = check.opt_inst_param(
            dagster_type, 'dagster_type', types.DagsterType, types.Any
        )

        if materializations is None and dagster_type is not None:
            materializations = dagster_type.default_materializations

        self.materializations = check.opt_list_param(
            materializations, 'materializations', of_type=MaterializationDefinition
        )

        self.expectations = check.opt_list_param(
            expectations, 'expectations', of_type=ExpectationDefinition
        )
        self.output_callback = check.opt_callable_param(output_callback, 'output_callback')
        self.description = check.opt_str_param(description, 'description')

    def materialization_of_type(self, name):
        for materialization in self.materializations:
            if materialization.name == name:
                return materialization

        check.failed('Did not find materialization {type}'.format(type=name))


# One or more inputs
# The core computation in the native kernel abstraction
# The output
class SolidDefinition:
    def __init__(self, name, inputs, transform_fn, output, description=None):
        self.name = check_valid_name(name)
        self.inputs = check.list_param(inputs, 'inputs', InputDefinition)
        self.output = check.inst_param(output, 'output', OutputDefinition)
        self.transform_fn = check.callable_param(transform_fn, 'transform')
        self.description = check.opt_str_param(description, 'description')

    # Notes to self

    # Input Definitions
    #   - DematerializationDefinitions
    #       - Arguments
    #       - Compute (args) => Value
    #   - Expectations
    #   - Dependency

    # Transform Definition
    #   - Function (inputs) => Value
    #   - Runtime Types (Inputs and Outputs)

    # Output Definition
    #   - MaterializationDefinitions
    #       - Arguments
    #       - Compute (value, args) => Result
    #   - Expectations

    @property
    def input_names(self):
        return [inp.name for inp in self.inputs]

    def has_input(self, name):
        check.str_param(name, 'name')
        for input_def in self.inputs:
            if input_def.name == name:
                return True
        return False

    def input_def_named(self, name):
        check.str_param(name, 'name')
        for input_def in self.inputs:
            if input_def.name == name:
                return input_def

        check.failed('input {name} not found'.format(name=name))


class __ArgumentValueSentinel:
    pass


NO_DEFAULT_PROVIDED = __ArgumentValueSentinel


class ArgumentDefinition:
    def __init__(
        self, dagster_type, default_value=NO_DEFAULT_PROVIDED, is_optional=False, description=None
    ):
        if not is_optional:
            check.param_invariant(
                default_value == NO_DEFAULT_PROVIDED,
                'default_value',
                'required arguments should not specify default values',
            )

        self.dagster_type = check.inst_param(dagster_type, 'dagster_type', types.DagsterType)
        self.description = check.opt_str_param(description, 'description')
        self.is_optional = check.bool_param(is_optional, 'is_optional')
        self.default_value = default_value

    @property
    def default_provided(self):
        return self.default_value != NO_DEFAULT_PROVIDED



import pytest

from dagster import (check, config)

from dagster.core.definitions import (SolidDefinition, OutputDefinition, PipelineDefinition)
from dagster.core.graph import (create_adjacency_lists, SolidGraph)
from dagster.core.execution import (
    execute_pipeline_iterator,
    execute_pipeline_iterator_in_memory,
    ExecutionContext,
    SolidExecutionResult,
)

from dagster.utils.compatability import create_custom_source_input

# protected members
# pylint: disable=W0212


def create_test_context():
    return ExecutionContext()


def _default_passthrough_transform(*args, **kwargs):
    check.invariant(not args, 'There should be no positional args')
    return list(kwargs.values())[0]


def create_dep_input_fn(name):
    return lambda context, arg_dict: {name: 'input_set'}


def create_solid_with_deps(name, *solid_deps):
    # def throw_input_fn(arg_dict):
    #     return [{name: 'input_set'}]

    inputs = [
        create_custom_source_input(
            solid_dep.name,
            source_fn=create_dep_input_fn(solid_dep.name),
            argument_def_dict={},
            depends_on=solid_dep,
        ) for solid_dep in solid_deps
    ]

    def dep_transform(_context, args):
        passed_rows = list(args.values())[0]
        passed_rows.append({name: 'transform_called'})
        #return copy.deepcopy(passed_rows)
        return passed_rows

    return SolidDefinition(
        name=name,
        inputs=inputs,
        transform_fn=dep_transform,
        output=OutputDefinition(),
    )


def create_root_solid(name):
    input_name = name + '_input'
    inp = create_custom_source_input(
        input_name,
        source_fn=lambda context, arg_dict: [{input_name: 'input_set'}],
        argument_def_dict={},
    )

    def root_transform(_context, args):
        passed_rows = list(args.values())[0]
        passed_rows.append({name: 'transform_called'})
        #return copy.deepcopy(passed_rows)
        return passed_rows

    return SolidDefinition(
        name=name,
        inputs=[inp],
        transform_fn=root_transform,
        output=OutputDefinition(),
    )


def test_empty_adjaceny_lists():
    forward_edges, backwards_edges = create_adjacency_lists([create_root_solid('a_node')])
    assert forward_edges == {'a_node': set()}
    assert backwards_edges == {'a_node': set()}


def test_single_dep_adjacency_lists():
    # A <-- B
    node_a = create_root_solid('A')
    node_b = create_solid_with_deps('B', node_a)

    forward_edges, backwards_edges = create_adjacency_lists([node_a, node_b])

    assert forward_edges == {'A': {'B'}, 'B': set()}
    assert backwards_edges == {'B': {'A'}, 'A': set()}


def test_diamond_deps_adjaceny_lists():
    # A <-- (B, C) <-- D

    node_a = create_root_solid('A')
    node_b = create_solid_with_deps('B', node_a)
    node_c = create_solid_with_deps('C', node_a)
    node_d = create_solid_with_deps('D', node_b, node_c)

    forward_edges, backwards_edges = create_adjacency_lists([node_a, node_b, node_c, node_d])
    assert forward_edges == {'A': {'B', 'C'}, 'B': {'D'}, 'C': {'D'}, 'D': set()}
    assert backwards_edges == {'D': {'B', 'C'}, 'B': {'A'}, 'C': {'A'}, 'A': set()}


def test_disconnected_graphs_adjaceny_lists():
    # A <-- B
    # C <-- D
    node_a = create_root_solid('A')
    node_b = create_solid_with_deps('B', node_a)

    node_c = create_root_solid('C')
    node_d = create_solid_with_deps('D', node_c)

    forward_edges, backwards_edges = create_adjacency_lists([node_a, node_b, node_c, node_d])
    assert forward_edges == {'A': {'B'}, 'B': set(), 'C': {'D'}, 'D': set()}
    assert backwards_edges == {'B': {'A'}, 'A': set(), 'D': {'C'}, 'C': set()}


def create_diamond_graph():
    node_a = create_root_solid('A')
    node_b = create_solid_with_deps('B', node_a)
    node_c = create_solid_with_deps('C', node_a)
    node_d = create_solid_with_deps('D', node_b, node_c)

    return SolidGraph([node_d, node_c, node_b, node_a])


def test_diamond_toposort():
    graph = create_diamond_graph()
    assert graph.topological_order == ['A', 'B', 'C', 'D']


def test_single_node_unprovided_inputs():
    node_a = create_root_solid('A')
    solid_graph = SolidGraph(solids=[node_a])
    assert solid_graph.compute_unprovided_inputs('A', ['A_input']) == set()
    assert solid_graph.compute_unprovided_inputs('A', []) == set(['A_input'])


def test_diamond_toposort_unprovided_inputs():
    solid_graph = create_diamond_graph()

    # no inputs
    assert solid_graph.compute_unprovided_inputs('A', []) == set(['A_input'])
    assert solid_graph.compute_unprovided_inputs('B', []) == set(['A_input'])
    assert solid_graph.compute_unprovided_inputs('C', []) == set(['A_input'])
    assert solid_graph.compute_unprovided_inputs('D', []) == set(['A_input'])

    # root input
    assert solid_graph.compute_unprovided_inputs('A', ['A_input']) == set()
    assert solid_graph.compute_unprovided_inputs('B', ['A_input']) == set()
    assert solid_graph.compute_unprovided_inputs('C', ['A_input']) == set()
    assert solid_graph.compute_unprovided_inputs('D', ['A_input']) == set()

    # immediate input
    assert solid_graph.compute_unprovided_inputs('A', ['A_input']) == set()
    assert solid_graph.compute_unprovided_inputs('B', ['A']) == set()
    assert solid_graph.compute_unprovided_inputs('C', ['A']) == set()
    assert solid_graph.compute_unprovided_inputs('D', ['B', 'C']) == set()

    # mixed satisified inputs
    assert solid_graph.compute_unprovided_inputs('D', ['A_input', 'C']) == set()
    assert solid_graph.compute_unprovided_inputs('D', ['B', 'A_input']) == set()

    # mixed unsatisifed inputs
    assert solid_graph.compute_unprovided_inputs('D', ['C']) == set(['A_input'])
    assert solid_graph.compute_unprovided_inputs('D', ['B']) == set(['A_input'])


def test_unprovided_input_param_invariants():
    node_a = create_root_solid('A')
    solid_graph = SolidGraph(solids=[node_a])

    with pytest.raises(check.ParameterCheckError):
        solid_graph.compute_unprovided_inputs('B', [])

    with pytest.raises(check.ParameterCheckError):
        solid_graph.compute_unprovided_inputs('A', ['no_input_here'])


def test_execution_subgraph_one_node():
    node_a = create_root_solid('A')
    solid_graph = SolidGraph(solids=[node_a])

    execution_graph = solid_graph.create_execution_subgraph(
        from_solids=['A'],
        to_solids=['A'],
    )
    assert execution_graph


def test_execution_graph_diamond():
    solid_graph = create_diamond_graph()

    full_execution_graph = solid_graph.create_execution_subgraph(
        from_solids=['A'],
        to_solids=['D'],
    )
    assert full_execution_graph.topological_order == ['A', 'B', 'C', 'D']
    assert len(full_execution_graph._solid_dict) == 4

    d_only_graph = solid_graph.create_execution_subgraph(
        from_solids=['D'],
        to_solids=['D'],
    )
    assert len(d_only_graph._solid_dict) == 1
    assert d_only_graph.topological_order == ['D']

    abc_graph = solid_graph.create_execution_subgraph(
        from_solids=['B', 'C'],
        to_solids=['D'],
    )
    assert len(abc_graph._solid_dict) == 3
    assert abc_graph.topological_order == ['B', 'C', 'D']


def input_set(name):
    return {name: 'input_set'}


def transform_called(name):
    return {name: 'transform_called'}


def assert_equivalent_results(left, right):
    check.inst_param(left, 'left', SolidExecutionResult)
    check.inst_param(right, 'right', SolidExecutionResult)

    assert left.success == right.success
    assert left.name == right.name
    assert left.solid.name == right.solid.name
    assert left.transformed_value == right.transformed_value


def assert_all_results_equivalent(expected_results, result_results):
    check.list_param(expected_results, 'expected_results', of_type=SolidExecutionResult)
    check.list_param(result_results, 'result_results', of_type=SolidExecutionResult)
    assert len(expected_results) == len(result_results)
    for expected, result in zip(expected_results, result_results):
        assert_equivalent_results(expected, result)


def test_pipeline_execution_graph_diamond():
    solid_graph = create_diamond_graph()
    pipeline = PipelineDefinition(solids=solid_graph.solids)
    environment = config.Environment(sources={'A': {'A_input': config.Source('CUSTOM', {})}})
    return _do_test(pipeline, lambda: execute_pipeline_iterator(
        pipeline,
        environment=environment,
    ))


def test_pipeline_execution_graph_diamond_in_memory():
    solid_graph = create_diamond_graph()
    pipeline = PipelineDefinition(solids=solid_graph.solids)
    input_values = {'A_input': [{'A_input': 'input_set'}]}
    return _do_test(pipeline, lambda: execute_pipeline_iterator_in_memory(
        ExecutionContext(),
        pipeline,
        input_values=input_values,
    ))


def _do_test(pipeline, do_execute_pipeline_iter):
    solid_graph = create_diamond_graph()

    pipeline = PipelineDefinition(solids=solid_graph.solids)

    results = list()

    for result in do_execute_pipeline_iter():
        results.append(result.copy())

    assert results[0].transformed_value[0] == input_set('A_input')
    assert results[0].transformed_value[1] == transform_called('A')

    assert results[0].transformed_value == [input_set('A_input'), transform_called('A')]

    expected_results = [
        SolidExecutionResult(
            success=True,
            solid=pipeline.solid_named('A'),
            transformed_value=[
                input_set('A_input'),
                transform_called('A'),
            ],
            exception=None,
        ),
        SolidExecutionResult(
            success=True,
            solid=pipeline.solid_named('B'),
            transformed_value=[
                input_set('A_input'),
                transform_called('A'),
                transform_called('B'),
            ],
            exception=None,
        ),
        SolidExecutionResult(
            success=True,
            solid=pipeline.solid_named('C'),
            transformed_value=[
                input_set('A_input'),
                transform_called('A'),
                transform_called('B'),
                transform_called('C'),
            ],
            exception=None,
        ),
        SolidExecutionResult(
            success=True,
            solid=pipeline.solid_named('D'),
            transformed_value=[
                input_set('A_input'),
                transform_called('A'),
                transform_called('B'),
                transform_called('C'),
                transform_called('D'),
            ],
            exception=None,
        ),
    ]

    assert_all_results_equivalent(expected_results, results)



import json

import pandas as pd
import sqlalchemy as sa

from dagster import check

if __name__ == '__main__':
    config_obj = json.load(open('config.json'))
    engine = sa.create_engine(
        'sqlite:///{dbname}.db'.format(dbname=check.str_elem(config_obj, 'dbname'))
    )
    for table in ['num_table', 'sum_table', 'sum_sq_table']:
        table_df = pd.read_sql_table(table, con=engine)
        print(table)
        print(table_df)



import sqlalchemy as sa

import dagster
from dagster.sqlalchemy_kernel.subquery_builder_experimental import (
    create_sql_solid,
    create_table_input_dependency,
    create_table_expression_input,
)


def create_engine_for_pipeline():
    engine = sa.create_engine('sqlite://')
    create_num_table(engine)
    return engine


def create_num_table(engine):
    metadata = sa.MetaData(engine)

    table = sa.Table(
        'num_table',
        metadata,
        sa.Column('num1', sa.Integer),
        sa.Column('num2', sa.Integer),
    )

    table.create()

    conn = engine.connect()

    conn.execute('''INSERT INTO num_table VALUES(1, 2)''')
    conn.execute('''INSERT INTO num_table VALUES(3, 4)''')


def define_pipeline():
    sum_table_solid = create_sql_solid(
        name='sum_table',
        inputs=[create_table_expression_input('num_table')],
        sql_text='SELECT num1, num2, num1 + num2 as sum FROM ({num_table})',
    )

    sum_sq_table_solid = create_sql_solid(
        name='sum_sq_table',
        inputs=[create_table_input_dependency(sum_table_solid)],
        sql_text='SELECT num1, num2, sum, sum * sum as sum_sq from ({sum_table})',
    )

    return dagster.PipelineDefinition(
        name='sql_hello_world', solids=[sum_table_solid, sum_sq_table_solid]
    )



from collections import (namedtuple, defaultdict)

from dagster import check


# lifted from https://bit.ly/2HcQAuv
class Materialization(namedtuple('MaterializationData', 'solid name args')):
    def __new__(cls, solid, name, args):
        return super(Materialization, cls).__new__(
            cls,
            solid=check.str_param(solid, 'solid'),
            name=check.str_param(name, 'name'),
            args=check.dict_param(args, 'args', key_type=str),
        )


class Context(namedtuple('ContextData', 'name args')):
    def __new__(cls, name, args):
        return super(Context, cls).__new__(
            cls, check.str_param(name, 'name'), check.dict_param(args, 'args', key_type=str)
        )


class Execution(namedtuple('ExecutionData', 'from_solids through_solids')):
    def __new__(cls, from_solids=None, through_solids=None):
        return super(Execution, cls).__new__(
            cls,
            check.opt_list_param(from_solids, 'from_solids', of_type=str),
            check.opt_list_param(through_solids, 'through_solids', of_type=str),
        )

    @staticmethod
    def single_solid(solid_name):
        check.str_param(solid_name, 'solid_name')
        return Execution(from_solids=[solid_name], through_solids=[solid_name])


class Environment(
    namedtuple('EnvironmentData', 'context sources materializations expectations, execution')
):
    def __new__(
        cls, sources, *, context=None, materializations=None, expectations=None, execution=None
    ):
        check.dict_param(sources, 'sources', key_type=str, value_type=dict)
        for _solid_name, source_dict in sources.items():
            check.dict_param(source_dict, 'source_dict', key_type=str, value_type=Source)

        check.opt_inst_param(context, 'context', Context)
        check.opt_inst_param(execution, 'execution', Execution)

        if context is None:
            context = Context(name='default', args={})

        if expectations is None:
            expectations = Expectations(evaluate=True)

        if execution is None:
            execution = Execution()

        return super(Environment, cls).__new__(
            cls,
            context=context,
            sources=sources,
            materializations=check.opt_list_param(
                materializations, 'materializations', of_type=Materialization
            ),
            expectations=expectations,
            execution=execution,
        )

    @staticmethod
    def empty():
        return Environment(sources={})


class Source(namedtuple('SourceData', 'name args')):
    def __new__(cls, name, args):
        return super(Source, cls).__new__(
            cls,
            name=check.str_param(name, 'name'),
            args=check.dict_param(args, 'args', key_type=str),
        )


class Expectations(namedtuple('ExpectationsData', 'evaluate')):
    def __new__(cls, evaluate):
        return super(Expectations, cls).__new__(
            cls, evaluate=check.bool_param(evaluate, 'evaluate')
        )


def _construct_context(yml_config_object):
    context_obj = check.opt_dict_elem(yml_config_object, 'context')
    if context_obj:
        return Context(check.str_elem(context_obj, 'name'), check.dict_elem(context_obj, 'args'))
    else:
        return None


def _construct_sources(yml_config_object):
    sources = defaultdict(dict)
    sources_obj = check.dict_elem(yml_config_object, 'sources')
    for solid_name, args_yml in sources_obj.items():
        for input_name, source_yml in args_yml.items():
            sources[solid_name][input_name] = Source(
                name=source_yml['name'], args=source_yml['args']
            )
    return sources


def _construct_materializations(yml_config_object):
    materializations = [
        Materialization(solid=m['solid'], name=m['name'], args=m['args'])
        for m in check.opt_list_elem(yml_config_object, 'materializations')
    ]
    return materializations


def _coerce_to_list(value):
    if value is None:
        return None
    elif isinstance(value, str):
        return [value]
    elif isinstance(value, list):
        return value

    check.invariant('should not get here')


def _construct_execution(yml_config_object):
    execution_obj = check.opt_dict_elem(yml_config_object, 'execution')
    if execution_obj is None:
        return None

    return Execution(
        from_solids=_coerce_to_list(execution_obj.get('from')),
        through_solids=_coerce_to_list(execution_obj.get('through')),
    )


def construct_environment(yml_config_object):
    check.dict_param(yml_config_object, 'yml_config_object')

    return Environment(
        sources=_construct_sources(yml_config_object),
        materializations=_construct_materializations(yml_config_object),
        context=_construct_context(yml_config_object),
        execution=_construct_execution(yml_config_object),
    )



from dagster.utils import has_context_argument


def test_has_context_variable():
    # pylint: disable=W0613

    def nope(_foo):
        pass

    def yup(context, _bar):
        pass

    assert not has_context_argument(nope)
    assert has_context_argument(yup)
    assert not has_context_argument(lambda: None)
    assert not has_context_argument(lambda bar: None)
    assert has_context_argument(lambda context: None)
    assert has_context_argument(lambda bar, context: None)



import pytest

from dagster.utils.indenting_printer import IndentingPrinter


class CollectingIndentingPrinter(IndentingPrinter):
    def __init__(self, *args, **kwargs):
        self.lines = []

        def _add_line(text):
            if str is not None:
                self.lines.append(text)

        super().__init__(printer=_add_line, *args, **kwargs)

    def result(self):
        return '\n'.join(self.lines)


def create_printer(*args, **kwargs):
    return CollectingIndentingPrinter(*args, **kwargs)


def test_basic_printer():
    printer = create_printer()
    printer.line('test')

    assert printer.result() == 'test'


def test_indent_printer():
    printer = create_printer()
    printer.line('test')
    with printer.with_indent():
        printer.line('test indent')

    assert printer.result() == '''test
  test indent'''


def test_parameterized_indent():
    printer = create_printer(indent_level=4)
    printer.line('test')
    with printer.with_indent():
        printer.line('test indent')

    assert printer.result() == '''test
    test indent'''


def test_bad_decrease_indent():
    printer = create_printer(indent_level=4)
    with pytest.raises(Exception):
        printer.decrease_indent()


def test_indent_printer_blank_line():
    printer = create_printer()
    printer.line('test')
    printer.blank_line()
    with printer.with_indent():
        printer.line('test indent')

    assert printer.result() == '''test

  test indent'''


def test_double_indent():
    printer = create_printer()
    printer.line('test')
    with printer.with_indent():
        printer.line('test indent')
        with printer.with_indent():
            printer.line('test double indent')

    assert printer.result() == '''test
  test indent
    test double indent'''


import pytest

from dagster import (
    ArgumentDefinition,
    ExecutionContext,
    OutputDefinition,
    PipelineDefinition,
    PipelineContextDefinition,
    config,
    execute_pipeline,
    solid,
    types,
    with_context,
)
from dagster.core.errors import (DagsterTypeError, DagsterInvariantViolationError)


def test_default_context():
    @solid(
        inputs=[],
        output=OutputDefinition(),
    )
    @with_context
    def default_context_transform(context):
        assert context.log_level == 'ERROR'

    pipeline = PipelineDefinition(solids=[default_context_transform])
    execute_pipeline(
        pipeline, environment=config.Environment(sources={}, context=config.Context('default', {}))
    )


def test_default_context_with_log_level():
    @solid(
        inputs=[],
        output=OutputDefinition(),
    )
    @with_context
    def default_context_transform(context):
        assert context.log_level == 'ERROR'

    pipeline = PipelineDefinition(solids=[default_context_transform])
    execute_pipeline(
        pipeline,
        environment=config.Environment(
            sources={}, context=config.Context('default', {'log_level': 'ERROR'})
        )
    )

    with pytest.raises(DagsterTypeError, message='Argument mismatch in context default'):
        execute_pipeline(
            pipeline,
            environment=config.Environment(
                sources={}, context=config.Context('default', {'log_level': 2})
            )
        )


def test_default_value():
    def _get_args_test_solid(arg_name, arg_value):
        @solid(
            inputs=[],
            output=OutputDefinition(),
        )
        @with_context
        def args_test(context):
            assert context.resources == {arg_name: arg_value}

        return args_test

    pipeline = PipelineDefinition(
        solids=[_get_args_test_solid('arg_one', 'heyo')],
        context_definitions={
            'custom_one':
            PipelineContextDefinition(
                argument_def_dict={
                    'arg_one':
                    ArgumentDefinition(
                        dagster_type=types.String,
                        is_optional=True,
                        default_value='heyo',
                    )
                },
                context_fn=lambda _pipeline, args: ExecutionContext(resources=args),
            ),
        }
    )

    execute_pipeline(
        pipeline,
        environment=config.Environment(sources={}, context=config.Context('custom_one', {}))
    )


def test_custom_contexts():
    @solid(
        inputs=[],
        output=OutputDefinition(),
    )
    @with_context
    def custom_context_transform(context):
        assert context.resources == {'arg_one': 'value_two'}

    pipeline = PipelineDefinition(
        solids=[custom_context_transform],
        context_definitions={
            'custom_one':
            PipelineContextDefinition(
                argument_def_dict={'arg_one': ArgumentDefinition(dagster_type=types.String)},
                context_fn=lambda _pipeline, args: ExecutionContext(resources=args),
            ),
            'custom_two':
            PipelineContextDefinition(
                argument_def_dict={'arg_one': ArgumentDefinition(dagster_type=types.String)},
                context_fn=lambda _pipeline, args: ExecutionContext(resources=args),
            )
        },
    )

    environment_one = config.Environment(
        sources={}, context=config.Context('custom_one', {'arg_one': 'value_two'})
    )

    execute_pipeline(pipeline, environment=environment_one)

    environment_two = config.Environment(
        sources={}, context=config.Context('custom_two', {'arg_one': 'value_two'})
    )

    execute_pipeline(pipeline, environment=environment_two)


def test_yield_context():
    events = []

    @solid(
        inputs=[],
        output=OutputDefinition(),
    )
    @with_context
    def custom_context_transform(context):
        assert context.resources == {'arg_one': 'value_two'}
        assert context._context_dict['foo'] == 'bar'  # pylint: disable=W0212
        events.append('during')

    def _yield_context(_pipeline, args):
        events.append('before')
        context = ExecutionContext(resources=args)
        with context.value('foo', 'bar'):
            yield context
        events.append('after')

    pipeline = PipelineDefinition(
        solids=[custom_context_transform],
        context_definitions={
            'custom_one':
            PipelineContextDefinition(
                argument_def_dict={'arg_one': ArgumentDefinition(dagster_type=types.String)},
                context_fn=_yield_context,
            ),
        }
    )

    environment_one = config.Environment(
        sources={}, context=config.Context('custom_one', {'arg_one': 'value_two'})
    )

    execute_pipeline(pipeline, environment=environment_one)

    assert events == ['before', 'during', 'after']


# TODO: reenable pending the ability to specific optional arguments
# https://github.com/dagster-io/dagster/issues/56
def test_invalid_context():
    @solid(
        inputs=[],
        output=OutputDefinition(),
    )
    def never_transform():
        raise Exception('should never execute')

    default_context_pipeline = PipelineDefinition(solids=[never_transform])

    environment_context_not_found = config.Environment(
        sources={}, context=config.Context('not_found', {})
    )

    with pytest.raises(DagsterInvariantViolationError, message='Context not_found not found'):
        execute_pipeline(
            default_context_pipeline,
            environment=environment_context_not_found,
            throw_on_error=True
        )

    environment_arg_name_mismatch = config.Environment(
        sources={}, context=config.Context('default', {'unexpected': 'value'})
    )

    with pytest.raises(DagsterTypeError, message='Argument mismatch in context default'):
        execute_pipeline(
            default_context_pipeline,
            environment=environment_arg_name_mismatch,
            throw_on_error=True
        )

    with_argful_context_pipeline = PipelineDefinition(
        solids=[never_transform],
        context_definitions={
            'default':
            PipelineContextDefinition(
                argument_def_dict={'string_arg': ArgumentDefinition(types.String)},
                context_fn=lambda _pipeline, _args: _args
            )
        }
    )

    environment_no_args_error = config.Environment(
        sources={}, context=config.Context('default', {})
    )

    with pytest.raises(DagsterTypeError, message='Argument mismatch in context default'):
        execute_pipeline(
            with_argful_context_pipeline,
            environment=environment_no_args_error,
            throw_on_error=True
        )

    environment_type_mismatch_error = config.Environment(
        sources={}, context=config.Context('default', {'string_arg': 1})
    )

    with pytest.raises(DagsterTypeError, message='Argument mismatch in context default'):
        execute_pipeline(
            with_argful_context_pipeline,
            environment=environment_type_mismatch_error,
            throw_on_error=True
        )



import logging
import os
import re
import textwrap

import click
import yaml

import dagster
from dagster import check
from dagster.core.execution import (DagsterExecutionFailureReason, execute_pipeline_iterator)
from dagster.graphviz import build_graphviz_graph
from dagster.utils.indenting_printer import IndentingPrinter

from .context import Config


def create_pipeline_cli():
    group = click.Group(name="pipeline")
    group.add_command(list_command)
    group.add_command(print_command)
    group.add_command(graphviz_command)
    group.add_command(execute_command)
    return group


@click.command(name='list', help="list")
@Config.pass_object
def list_command(config):
    pipeline_configs = config.create_pipelines()

    for pipeline_config in pipeline_configs:
        pipeline = pipeline_config.pipeline
        click.echo('Pipeline: {name}'.format(name=pipeline.name))
        if pipeline.description:
            click.echo('Description:')
            click.echo(format_description(pipeline.description, indent=' ' * 4))
        click.echo('Solids: (Execution Order)')
        for solid in pipeline.solid_graph.topological_solids:
            click.echo('    ' + solid.name)
        click.echo('*************')


def format_description(desc, indent):
    check.str_param(desc, 'desc')
    check.str_param(indent, 'indent')
    desc = re.sub(r'\s+', ' ', desc)
    dedented = textwrap.dedent(desc)
    wrapper = textwrap.TextWrapper(initial_indent='', subsequent_indent=indent)
    filled = wrapper.fill(dedented)
    return filled


def set_pipeline(ctx, _arg, value):
    ctx.params['pipeline_config'] = ctx.find_object(Config).get_pipeline(value)


def pipeline_name_argument(f):
    return click.argument('pipeline_name', callback=set_pipeline, expose_value=False)(f)


@click.command(name='print', help="print <<pipeline_name>>")
@pipeline_name_argument
def print_command(pipeline_config):
    print_pipeline(pipeline_config.pipeline, full=True, print_fn=click.echo)


def print_pipeline(pipeline, full, print_fn):
    check.inst_param(pipeline, 'pipeline', dagster.PipelineDefinition)
    check.bool_param(full, 'full')
    check.callable_param(print_fn, 'print_fn')

    printer = IndentingPrinter(indent_level=2, printer=print_fn)
    printer.line('Pipeline: {name}'.format(name=pipeline.name))
    print_description(printer, pipeline.description)

    if not full:
        return

    with printer.with_indent():
        printer.line('Context Definitions:')

        with printer.with_indent():

            for context_name, context_definition in pipeline.context_definitions.items():
                print_context_definition(printer, context_name, context_definition)

    printer.line('Solids:')
    for solid in pipeline.solids:
        with printer.with_indent():
            print_solid(printer, solid)


def print_description(printer, desc):
    with printer.with_indent():
        if desc:
            printer.line('Description:')
            with printer.with_indent():
                printer.line(format_description(desc, printer.current_indent_str))


def print_context_definition(printer, context_name, context_definition):
    printer.line('Name: {context_name}'.format(context_name=context_name))

    print_description(printer, context_definition.description)

    printer.line('Args:')

    with printer.with_indent():
        for arg_name, arg_def in context_definition.argument_def_dict.items():
            printer.line('Arg: {name}'.format(name=arg_name))
            with printer.with_indent():
                printer.line('Type: {arg_type}'.format(arg_type=arg_def.dagster_type.name))

                print_description(printer, arg_def.description)


def print_solid(printer, solid):
    printer.line('Solid: {name}'.format(name=solid.name))

    with printer.with_indent():
        print_inputs(printer, solid)

        printer.line('Output:')

        if solid.output.materializations:
            printer.line('Materializations:')
            for materialization_def in solid.output.materializations:
                arg_list = format_argument_dict(materialization_def.argument_def_dict)
                with printer.with_indent():
                    printer.line(
                        '{name}({arg_list})'.format(
                            name=materialization_def.name, arg_list=arg_list
                        )
                    )


def print_inputs(printer, solid):
    printer.line('Inputs:')
    for input_def in solid.inputs:
        with printer.with_indent():
            if input_def.depends_on:
                printer.line(
                    'Input: {name} (depends on {dep_name})'.format(
                        name=input_def.name, dep_name=input_def.depends_on.name
                    )
                )
            else:
                printer.line('Input: {name}'.format(name=input_def.name))

            if input_def.sources:
                print_sources(printer, input_def.sources)


def print_sources(printer, sources):
    with printer.with_indent():
        printer.line('Sources:')
        with printer.with_indent():
            for source_def in sources:
                arg_list = format_argument_dict(source_def.argument_def_dict)
                printer.line(
                    '{input_name}({arg_list})'.format(
                        input_name=source_def.source_type,
                        arg_list=arg_list,
                    )
                )


def format_argument_dict(arg_def_dict):
    return ', '.join(
        [
            '{name}: {type}'.format(name=name, type=arg_def.dagster_type.name)
            for name, arg_def in arg_def_dict.items()
        ]
    )


@click.command(name='graphviz', help="graphviz <<pipeline_name>>")
@pipeline_name_argument
def graphviz_command(pipeline_config):
    build_graphviz_graph(pipeline_config.pipeline).view(cleanup=True)


def get_default_config_for_pipeline():
    ctx = click.get_current_context()
    pipeline_config = ctx.params['pipeline_config']
    module_path = os.path.dirname(pipeline_config.module.__file__)
    return os.path.join(module_path, 'env.yml')


LOGGING_DICT = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARN': logging.WARN,
    'ERROR': logging.ERROR,
    'CRITICAL': logging.CRITICAL,
}


def load_yaml_from_path(path):
    check.str_param(path, 'path')
    with open(path, 'r') as ff:
        return yaml.load(ff)


@click.command(name='execute', help="execute <<pipeline_name>>")
@pipeline_name_argument
@click.option(
    '-e',
    '--env',
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
    ),
    default=get_default_config_for_pipeline,
    help="Path to environment file. Defaults to ./PIPELINE_DIR/env.yml."
)
def execute_command(pipeline_config, env):
    do_execute_command(pipeline_config.pipeline, env, print)


def do_execute_command(pipeline, env, printer):
    check.inst_param(pipeline, 'pipeline', dagster.PipelineDefinition)
    check.str_param(env, 'env')
    check.callable_param(printer, 'printer')

    env_config = load_yaml_from_path(env)

    environment = dagster.config.construct_environment(env_config)

    pipeline_iter = execute_pipeline_iterator(pipeline, environment)

    process_results_for_console(pipeline_iter, printer)


def process_results_for_console(pipeline_iter, printer):
    results = []

    for result in pipeline_iter:
        if not result.success:
            if result.reason == DagsterExecutionFailureReason.USER_CODE_ERROR:
                raise result.user_exception
            elif result.reason == DagsterExecutionFailureReason.EXPECTATION_FAILURE:
                for expectation_result in result.failed_expectation_results:
                    result.context.error(expectation_result.message, solid=result.solid.name)
                click_context = click.get_current_context()
                click_context.exit(1)
        results.append(result)

    print_metrics_to_console(results, printer)


def print_metrics_to_console(results, printer):
    for result in results:
        context = result.context
        metrics_of_solid = list(context.metrics_matching_context({'solid': result.name}))

        printer('Metrics for {name}'.format(name=result.name))

        for input_def in result.solid.inputs:
            metrics_for_input = list(
                context.metrics_covering_context({
                    'solid': result.name,
                    'input': input_def.name,
                })
            )
            if metrics_for_input:
                printer('    Input {input_name}'.format(input_name=input_def.name))
                for metric in metrics_for_input:
                    printer(
                        '{indent}{metric_name}: {value}'.format(
                            indent=' ' * 8, metric_name=metric.metric_name, value=metric.value
                        )
                    )

        for metric in metrics_of_solid:
            printer(
                '{indent}{metric_name}: {value}'.format(
                    indent=' ' * 4, metric_name=metric.metric_name, value=metric.value
                )
            )



import os

from dagster import check

from dagster.core import types

from dagster.core.definitions import (
    SolidDefinition,
    InputDefinition,
    SourceDefinition,
    ArgumentDefinition,
    OutputDefinition,
)

from dagster.utils.compatability import create_single_materialization_output

from dagster.sqlalchemy_kernel import execute_sql_text_on_context


class DagsterSqlExpression:
    @property
    def from_target(self):
        check.not_implemented('must implemented in subclass')


class DagsterSqlQueryExpression(DagsterSqlExpression):
    def __init__(self, subquery_text):
        super().__init__()
        self._subquery_text = check.str_param(subquery_text, 'subquery_text')

    @property
    def query_text(self):
        return self._subquery_text

    @property
    def from_target(self):
        return f'({self._subquery_text})'


class DagsterSqlTableExpression(DagsterSqlExpression):
    def __init__(self, table_name):
        super().__init__()
        self._table_name = check.str_param(table_name, 'table_name')

    @property
    def query_text(self):
        check.not_implemented('table cannot be a standalone query')

    @property
    def from_target(self):
        return self._table_name


def create_table_output():
    def materialization_fn(context, arg_dict, sql_expr):
        check.inst_param(sql_expr, 'sql_expr', DagsterSqlExpression)
        check.dict_param(arg_dict, 'arg_dict')

        output_table_name = check.str_elem(arg_dict, 'table_name')
        total_sql = '''CREATE TABLE {output_table_name} AS {query_text}'''.format(
            output_table_name=output_table_name, query_text=sql_expr.query_text
        )
        context.resources.sa.engine.connect().execute(total_sql)

    return create_single_materialization_output(
        name='CREATE',
        materialization_fn=materialization_fn,
        argument_def_dict={'table_name': ArgumentDefinition(types.String)}
    )


def _table_name_read_fn(_context, arg_dict):
    check.dict_param(arg_dict, 'arg_dict')

    table_name = check.str_elem(arg_dict, 'table_name')
    # probably verify that the table name exists?
    return DagsterSqlTableExpression(table_name)


def _table_name_source():
    return SourceDefinition(
        source_type='TABLENAME',
        source_fn=_table_name_read_fn,
        argument_def_dict={'table_name': ArgumentDefinition(types.String)},
    )


def create_table_expression_input(name):
    check.str_param(name, 'name')
    return InputDefinition(name=name, sources=[_table_name_source()])


def create_table_input_dependency(solid):
    check.inst_param(solid, 'solid', SolidDefinition)

    return InputDefinition(name=solid.name, sources=[_table_name_source()], depends_on=solid)


def create_sql_transform(sql_text):
    def transform_fn(_context, args):
        sql_texts = {}
        for name, sql_expr in args.items():
            sql_texts[name] = sql_expr.from_target

        return DagsterSqlQueryExpression(sql_text.format(**sql_texts))

    return transform_fn


def create_sql_solid(name, inputs, sql_text):
    check.str_param(name, 'name')
    check.list_param(inputs, 'inputs', of_type=InputDefinition)
    check.str_param(sql_text, 'sql_text')

    return SolidDefinition(
        name,
        inputs=inputs,
        transform_fn=create_sql_transform(sql_text),
        output=create_table_output(),
    )


def _create_sql_alchemy_transform_fn(sql_text):
    check.str_param(sql_text, 'sql_text')

    def transform_fn(context, _args):
        return execute_sql_text_on_context(context, sql_text)

    return transform_fn


def create_sql_statement_solid(name, sql_text, inputs=None):
    check.str_param(name, 'name')
    check.str_param(sql_text, 'sql_text')
    check.opt_list_param(inputs, 'inputs', of_type=InputDefinition)

    if inputs is None:
        inputs = []

    return SolidDefinition(
        name=name,
        transform_fn=_create_sql_alchemy_transform_fn(sql_text),
        inputs=inputs,
        output=OutputDefinition()
    )


def sql_file_solid(path, inputs=None):
    check.str_param(path, 'path')
    check.opt_list_param(inputs, 'inputs', of_type=InputDefinition)

    basename = os.path.basename(path)
    name = os.path.splitext(basename)[0]

    with open(path, 'r') as ff:
        return create_sql_statement_solid(name, ff.read(), inputs)
